# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/6/23 15:09
@author: ZhangYundi
@email: yundi.xxii@outlook.com
@description: 
---------------------------------------------
"""

import abc
import mlflow


class BaseIncrementalTrainer(abc.ABC):
    """
    支持增量训练和mlflow管理的通用基础类

    该抽象基类定义了所有增量训练模型必须实现的核心接口，包括数据集构建(`create_dataset`)、训练流程(`fit`).
    """

    def __init__(self,
                 experiment_name: str, ):
        self.experiment_name = experiment_name
        mlflow.set_experiment(experiment_name)
        self.model = None
        self.input_example = None

    @abc.abstractmethod
    def create_dataset(self, data, features: list[str], target: str):
        """
        创建模型特定的数据集对象。

        子类需要实现此方法以将原始数据转换为模型可以接受的输入格式(如`lightgbm.Dataset`)

        Parameters
        ----------
        data: polars.DataFrame
            原始数据集，通常包含特征列和目标列
        features: list[str]
            特征列名列表
        target: str
            目标列名

        Returns
        -------
        object
            模型所需要的数据集对象(如`lightgbm.Dataset`)
        """
        pass

    @abc.abstractmethod
    def fit(self,
            train_data,
            valid_data=None,
            init_model=None,
            evals_result: dict | None = None,
            **model_params):
        """
        执行模型训练

        子类应该实现此方法以完成具体的训练逻辑，支持增量训练

        Parameters
        ----------
        train_data: object
            训练集对象，由`create_dataset`生成
        valid_data: object
            验证集对象，由`create_dataset`生成
        init_model: object, optional
            初始模型或者check_point.uri, 用于增量训练或者断点训练, 默认None
        evals_result: dict, optional
            用于存储评估结果的字典，默认None
        **model_params: dict
            模型超参

        Returns
        -------
        object
            训练完成的模型对象
        """
        pass

    def save_model(self, model, name: str):
        """保存模型到MLflow"""
        from mlflow.models.signature import infer_signature
        from lightgbm import Booster
        from torch.nn import Module
        signature = infer_signature(self.input_example)
        if isinstance(model, Booster):
            return mlflow.lightgbm.log_model(model, name=name, signature=signature, input_example=self.input_example)
        elif isinstance(model, Module):
            return mlflow.pytorch.log_model(model, name=name, signature=signature, input_example=self.input_example)
        else:
            raise ValueError("Unsupported model type")



#
# Copyright 2016 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import math
from math import pow
from sys import float_info

import numpy as np
import pandas as pd
from scipy import stats, optimize
from six import iteritems

from .periods import ANNUALIZATION_FACTORS, APPROX_BDAYS_PER_YEAR
from .periods import DAILY, WEEKLY, MONTHLY, QUARTERLY, YEARLY
from .utils import up, down, roll, rolling_window


def _create_unary_vectorized_roll_function(function):
    def unary_vectorized_roll(arr, window, out=None, **kwargs):
        """
        Computes the {human_readable} measure over a rolling window.

        Parameters
        ----------
        arr : array-like
            The array to compute the rolling {human_readable} over.
        window : int
            Size of the rolling window in terms of the periodicity of the data.
        out : array-like, optional
            Array to use as output buffer.
            If not passed, a new array will be created.
        **kwargs
            Forwarded to :func:`~empyrical.{name}`.

        Returns
        -------
        rolling_{name} : array-like
            The rolling {human_readable}.
        """
        allocated_output = out is None

        if len(arr):
            out = function(
                rolling_window(_flatten(arr), min(len(arr), window)).T,
                out=out,
                **kwargs
            )
        else:
            out = np.empty(0, dtype='float64')

        if allocated_output and isinstance(arr, pd.Series):
            out = pd.Series(out, index=arr.index[-len(out):])

        return out

    unary_vectorized_roll.__doc__ = unary_vectorized_roll.__doc__.format(
        name=function.__name__,
        human_readable=function.__name__.replace('_', ' '),
    )

    return unary_vectorized_roll


def _create_binary_vectorized_roll_function(function):
    def binary_vectorized_roll(lhs, rhs, window, out=None, **kwargs):
        """
        Computes the {human_readable} measure over a rolling window.

        Parameters
        ----------
        lhs : array-like
            The first array to pass to the rolling {human_readable}.
        rhs : array-like
            The second array to pass to the rolling {human_readable}.
        window : int
            Size of the rolling window in terms of the periodicity of the data.
        out : array-like, optional
            Array to use as output buffer.
            If not passed, a new array will be created.
        **kwargs
            Forwarded to :func:`~empyrical.{name}`.

        Returns
        -------
        rolling_{name} : array-like
            The rolling {human_readable}.
        """
        allocated_output = out is None

        if window >= 1 and len(lhs) and len(rhs):
            out = function(
                rolling_window(_flatten(lhs), min(len(lhs), window)).T,
                rolling_window(_flatten(rhs), min(len(rhs), window)).T,
                out=out,
                **kwargs
            )
        elif allocated_output:
            out = np.empty(0, dtype='float64')
        else:
            out[()] = np.nan

        if allocated_output:
            if out.ndim == 1 and isinstance(lhs, pd.Series):
                out = pd.Series(out, index=lhs.index[-len(out):])
            elif out.ndim == 2 and isinstance(lhs, pd.Series):
                out = pd.DataFrame(out, index=lhs.index[-len(out):])
        return out

    binary_vectorized_roll.__doc__ = binary_vectorized_roll.__doc__.format(
        name=function.__name__,
        human_readable=function.__name__.replace('_', ' '),
    )

    return binary_vectorized_roll


def _flatten(arr):
    return arr if not isinstance(arr, pd.Series) else arr.values


def _adjust_returns(returns, adjustment_factor):
    """
    Returns the returns series adjusted by adjustment_factor. Optimizes for the
    case of adjustment_factor being 0 by returning returns itself, not a copy!

    Parameters
    ----------
    returns : pd.Series or np.ndarray
    adjustment_factor : pd.Series or np.ndarray or float or int

    Returns
    -------
    adjusted_returns : array-like
    """
    if isinstance(adjustment_factor, (float, int)) and adjustment_factor == 0:
        return returns
    return returns - adjustment_factor


def annualization_factor(period, annualization):
    """
    Return annualization factor from period entered or if a custom
    value is passed in.

    Parameters
    ----------
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.

    Returns
    -------
    annualization_factor : float
    """
    if annualization is None:
        try:
            factor = ANNUALIZATION_FACTORS[period]
        except KeyError:
            raise ValueError(
                "Period cannot be '{}'. "
                "Can be '{}'.".format(
                    period, "', '".join(ANNUALIZATION_FACTORS.keys())
                )
            )
    else:
        factor = annualization
    return factor


def simple_returns(prices):
    """
    Compute simple returns from a timeseries of prices.

    Parameters
    ----------
    prices : pd.Series, pd.DataFrame or np.ndarray
        Prices of assets in wide-format, with assets as columns,
        and indexed by datetimes.

    Returns
    -------
    returns : array-like
        Returns of assets in wide-format, with assets as columns,
        and index coerced to be tz-aware.
    """
    if isinstance(prices, (pd.DataFrame, pd.Series)):
        out = prices.pct_change().iloc[1:]
    else:
        # Assume np.ndarray
        out = np.diff(prices, axis=0)
        np.divide(out, prices[:-1], out=out)

    return out


def cum_returns(returns, starting_value=0, out=None):
    """
    Compute cumulative returns from simple returns.

    Parameters
    ----------
    returns : pd.Series, np.ndarray, or pd.DataFrame
        Returns of the strategy as a percentage, noncumulative.
         - Time series with decimal returns.
         - Example::

            2015-07-16   -0.012143
            2015-07-17    0.045350
            2015-07-20    0.030957
            2015-07-21    0.004902

         - Also accepts two dimensional data. In this case, each column is
           cumulated.

    starting_value : float, optional
       The starting returns.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    cumulative_returns : array-like
        Series of cumulative returns.
    """
    if len(returns) < 1:
        return returns.copy()

    nanmask = np.isnan(returns)
    if np.any(nanmask):
        returns = returns.copy()
        returns[nanmask] = 0

    allocated_output = out is None
    if allocated_output:
        out = np.empty_like(returns)

    np.add(returns, 1, out=out)
    out.cumprod(axis=0, out=out)

    if starting_value == 0:
        np.subtract(out, 1, out=out)
    else:
        np.multiply(out, starting_value, out=out)

    if allocated_output:
        if returns.ndim == 1 and isinstance(returns, pd.Series):
            out = pd.Series(out, index=returns.index)
        elif isinstance(returns, pd.DataFrame):
            out = pd.DataFrame(
                out, index=returns.index, columns=returns.columns,
            )

    return out


def cum_returns_final(returns, starting_value=0):
    """
    Compute total returns from simple returns.

    Parameters
    ----------
    returns : pd.DataFrame, pd.Series, or np.ndarray
       Noncumulative simple returns of one or more timeseries.
    starting_value : float, optional
       The starting returns.

    Returns
    -------
    total_returns : pd.Series, np.ndarray, or float
        If input is 1-dimensional (a Series or 1D numpy array), the result is a
        scalar.

        If input is 2-dimensional (a DataFrame or 2D numpy array), the result
        is a 1D array containing cumulative returns for each column of input.
    """
    if len(returns) == 0:
        return np.nan

    if isinstance(returns, pd.DataFrame):
        result = (returns + 1).prod()
    else:
        result = np.nanprod(returns + 1, axis=0)

    if starting_value == 0:
        result -= 1
    else:
        result *= starting_value

    return result


def aggregate_returns(returns, convert_to):
    """
    Aggregates returns by week, month, or year.

    Parameters
    ----------
    returns : pd.Series
       Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    convert_to : str
        Can be 'weekly', 'monthly', or 'yearly'.

    Returns
    -------
    aggregated_returns : pd.Series
    """

    def cumulate_returns(x):
        return cum_returns(x).iloc[-1]

    if convert_to == WEEKLY:
        grouping = [lambda x: x.year, lambda x: x.isocalendar()[1]]
    elif convert_to == MONTHLY:
        grouping = [lambda x: x.year, lambda x: x.month]
    elif convert_to == QUARTERLY:
        grouping = [lambda x: x.year, lambda x: int(math.ceil(x.month / 3.))]
    elif convert_to == YEARLY:
        grouping = [lambda x: x.year]
    else:
        raise ValueError(
            'convert_to must be {}, {} or {}'.format(WEEKLY, MONTHLY, YEARLY)
        )

    return returns.groupby(grouping).apply(cumulate_returns)


def max_drawdown(returns, out=None):
    """
    Determines the maximum drawdown of a strategy.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    max_drawdown : float

    Note
    -----
    See https://en.wikipedia.org/wiki/Drawdown_(economics) for more details.
    """
    allocated_output = out is None
    if allocated_output:
        out = np.empty(returns.shape[1:])

    returns_1d = returns.ndim == 1

    if len(returns) < 1:
        out[()] = np.nan
        if returns_1d:
            out = out.item()
        return out

    returns_array = np.asanyarray(returns)

    cumulative = np.empty(
        (returns.shape[0] + 1,) + returns.shape[1:],
        dtype='float64',
    )
    cumulative[0] = start = 100
    cum_returns(returns_array, starting_value=start, out=cumulative[1:])

    max_return = np.fmax.accumulate(cumulative, axis=0)

    np.nanmin((cumulative - max_return) / max_return, axis=0, out=out)
    if returns_1d:
        out = out.item()
    elif allocated_output and isinstance(returns, pd.DataFrame):
        out = pd.Series(out)

    return out


roll_max_drawdown = _create_unary_vectorized_roll_function(max_drawdown)


def annual_return(returns, period=DAILY, annualization=None):
    """
    Determines the mean annual growth rate of returns. This is equivilent
    to the compound annual growth rate.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Periodic returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.

    Returns
    -------
    annual_return : float
        Annual Return as CAGR (Compounded Annual Growth Rate).

    """

    if len(returns) < 1:
        return np.nan

    ann_factor = annualization_factor(period, annualization)
    num_years = len(returns) / ann_factor
    # Pass array to ensure index -1 looks up successfully.
    ending_value = cum_returns_final(returns, starting_value=1)

    return ending_value ** (1 / num_years) - 1


def cagr(returns, period=DAILY, annualization=None):
    """
    Compute compound annual growth rate. Alias function for
    :func:`~empyrical.stats.annual_return`

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.
        - See full explanation in :func:`~empyrical.stats.annual_return`.

    Returns
    -------
    cagr : float
        The CAGR value.

    """
    return annual_return(returns, period, annualization)


roll_cagr = _create_unary_vectorized_roll_function(cagr)


def annual_volatility(returns,
                      period=DAILY,
                      alpha=2.0,
                      annualization=None,
                      out=None):
    """
    Determines the annual volatility of a strategy.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Periodic returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    alpha : float, optional
        Scaling relation (Levy stability exponent).
    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    annual_volatility : float
    """
    allocated_output = out is None
    if allocated_output:
        out = np.empty(returns.shape[1:])

    returns_1d = returns.ndim == 1

    if len(returns) < 2:
        out[()] = np.nan
        if returns_1d:
            out = out.item()
        return out

    ann_factor = annualization_factor(period, annualization)
    np.nanstd(returns, ddof=1, axis=0, out=out)
    out = np.multiply(out, ann_factor ** (1.0 / alpha), out=out)
    if returns_1d:
        out = out.item()
    return out


roll_annual_volatility = _create_unary_vectorized_roll_function(
    annual_volatility,
)


def calmar_ratio(returns, period=DAILY, annualization=None):
    """
    Determines the Calmar ratio, or drawdown ratio, of a strategy.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.


    Returns
    -------
    calmar_ratio : float
        Calmar ratio (drawdown ratio) as float. Returns np.nan if there is no
        calmar ratio.

    Note
    -----
    See https://en.wikipedia.org/wiki/Calmar_ratio for more details.
    """

    max_dd = max_drawdown(returns=returns)
    if max_dd < 0:
        temp = annual_return(
            returns=returns,
            period=period,
            annualization=annualization
        ) / abs(max_dd)
    else:
        return np.nan

    if np.isinf(temp):
        return np.nan

    return temp


def omega_ratio(returns, risk_free=0.0, required_return=0.0,
                annualization=APPROX_BDAYS_PER_YEAR):
    """Determines the Omega ratio of a strategy.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    risk_free : int, float
        Constant risk-free return throughout the period
    required_return : float, optional
        Minimum acceptance return of the investor. Threshold over which to
        consider positive vs negative returns. It will be converted to a
        value appropriate for the period of the returns. E.g. An annual minimum
        acceptable return of 100 will translate to a minimum acceptable
        return of 0.018.
    annualization : int, optional
        Factor used to convert the required_return into a daily
        value. Enter 1 if no time period conversion is necessary.

    Returns
    -------
    omega_ratio : float

    Note
    -----
    See https://en.wikipedia.org/wiki/Omega_ratio for more details.

    """

    if len(returns) < 2:
        return np.nan

    if annualization == 1:
        return_threshold = required_return
    elif required_return <= -1:
        return np.nan
    else:
        return_threshold = (1 + required_return) ** \
                           (1. / annualization) - 1

    returns_less_thresh = returns - risk_free - return_threshold

    numer = sum(returns_less_thresh[returns_less_thresh > 0.0])
    denom = -1.0 * sum(returns_less_thresh[returns_less_thresh < 0.0])

    if denom > 0.0:
        return numer / denom
    else:
        return np.nan


def sharpe_ratio(returns,
                 risk_free=0,
                 period=DAILY,
                 annualization=None,
                 out=None):
    """
    Determines the Sharpe ratio of a strategy.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    risk_free : int, float
        Constant daily risk-free return throughout the period.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    sharpe_ratio : float
        nan if insufficient length of returns or if if adjusted returns are 0.

    Note
    -----
    See https://en.wikipedia.org/wiki/Sharpe_ratio for more details.

    """
    allocated_output = out is None
    if allocated_output:
        out = np.empty(returns.shape[1:])

    return_1d = returns.ndim == 1

    if len(returns) < 2:
        out[()] = np.nan
        if return_1d:
            out = out.item()
        return out

    returns_risk_adj = np.asanyarray(_adjust_returns(returns, risk_free))
    ann_factor = annualization_factor(period, annualization)

    np.multiply(
        np.divide(
            np.nanmean(returns_risk_adj, axis=0),
            np.nanstd(returns_risk_adj, ddof=1, axis=0),
            out=out,
        ),
        np.sqrt(ann_factor),
        out=out,
    )
    if return_1d:
        out = out.item()

    return out


roll_sharpe_ratio = _create_unary_vectorized_roll_function(sharpe_ratio)


def sortino_ratio(returns,
                  required_return=0,
                  period=DAILY,
                  annualization=None,
                  out=None,
                  _downside_risk=None):
    """
    Determines the Sortino ratio of a strategy.

    Parameters
    ----------
    returns : pd.Series or np.ndarray or pd.DataFrame
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    required_return: float / series
        minimum acceptable return
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.
    _downside_risk : float, optional
        The downside risk of the given inputs, if known. Will be calculated if
        not provided.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    sortino_ratio : float or pd.Series

        depends on input type
        series ==> float
        DataFrame ==> pd.Series

    Note
    -----
    See `<https://www.sunrisecapital.com/wp-content/uploads/2014/06/Futures_
    Mag_Sortino_0213.pdf>`__ for more details.

    """
    allocated_output = out is None
    if allocated_output:
        out = np.empty(returns.shape[1:])

    return_1d = returns.ndim == 1

    if len(returns) < 2:
        out[()] = np.nan
        if return_1d:
            out = out.item()
        return out

    adj_returns = np.asanyarray(_adjust_returns(returns, required_return))

    ann_factor = annualization_factor(period, annualization)

    average_annual_return = np.nanmean(adj_returns, axis=0) * ann_factor
    annualized_downside_risk = (
        _downside_risk
        if _downside_risk is not None else
        downside_risk(returns, required_return, period, annualization)
    )
    np.divide(average_annual_return, annualized_downside_risk, out=out)
    if return_1d:
        out = out.item()
    elif isinstance(returns, pd.DataFrame):
        out = pd.Series(out)

    return out


roll_sortino_ratio = _create_unary_vectorized_roll_function(sortino_ratio)


def downside_risk(returns,
                  required_return=0,
                  period=DAILY,
                  annualization=None,
                  out=None):
    """
    Determines the downside deviation below a threshold

    Parameters
    ----------
    returns : pd.Series or np.ndarray or pd.DataFrame
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    required_return: float / series
        minimum acceptable return
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    downside_deviation : float or pd.Series
        depends on input type
        series ==> float
        DataFrame ==> pd.Series

    Note
    -----
    See `<https://www.sunrisecapital.com/wp-content/uploads/2014/06/Futures_
    Mag_Sortino_0213.pdf>`__ for more details, specifically why using the
    standard deviation of the negative returns is not correct.
    """
    allocated_output = out is None
    if allocated_output:
        out = np.empty(returns.shape[1:])

    returns_1d = returns.ndim == 1

    if len(returns) < 1:
        out[()] = np.nan
        if returns_1d:
            out = out.item()
        return out

    ann_factor = annualization_factor(period, annualization)

    downside_diff = np.clip(
        _adjust_returns(
            np.asanyarray(returns),
            np.asanyarray(required_return),
        ),
        -np.inf,
        0,
    )

    np.square(downside_diff, out=downside_diff)
    np.nanmean(downside_diff, axis=0, out=out)
    np.sqrt(out, out=out)
    np.multiply(out, np.sqrt(ann_factor), out=out)

    if returns_1d:
        out = out.item()
    elif isinstance(returns, pd.DataFrame):
        out = pd.Series(out, index=returns.columns)
    return out


roll_downsize_risk = _create_unary_vectorized_roll_function(downside_risk)


def excess_sharpe(returns, factor_returns, out=None):
    """
    Determines the Excess Sharpe of a strategy.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns: float / series
        Benchmark return to compare returns against.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    excess_sharpe : float

    Note
    -----
    The excess Sharpe is a simplified Information Ratio that uses
    tracking error rather than "active risk" as the denominator.
    """
    allocated_output = out is None
    if allocated_output:
        out = np.empty(returns.shape[1:])

    returns_1d = returns.ndim == 1

    if len(returns) < 2:
        out[()] = np.nan
        if returns_1d:
            out = out.item()
        return out

    active_return = _adjust_returns(returns, factor_returns)
    tracking_error = np.nan_to_num(np.nanstd(active_return, ddof=1, axis=0))

    out = np.divide(
        np.nanmean(active_return, axis=0, out=out),
        tracking_error,
        out=out,
    )
    if returns_1d:
        out = out.item()
    return out


roll_excess_sharpe = _create_binary_vectorized_roll_function(excess_sharpe)


def _to_pandas(ob):
    """Convert an array-like to a pandas object.

    Parameters
    ----------
    ob : array-like
        The object to convert.

    Returns
    -------
    pandas_structure : pd.Series or pd.DataFrame
        The correct structure based on the dimensionality of the data.
    """
    if isinstance(ob, (pd.Series, pd.DataFrame)):
        return ob

    if ob.ndim == 1:
        return pd.Series(ob)
    elif ob.ndim == 2:
        return pd.DataFrame(ob)
    else:
        raise ValueError(
            'cannot convert array of dim > 2 to a pandas structure',
        )


def _aligned_series(*many_series):
    """
    Return a new list of series containing the data in the input series, but
    with their indices aligned. NaNs will be filled in for missing values.

    Parameters
    ----------
    *many_series
        The series to align.

    Returns
    -------
    aligned_series : iterable[array-like]
        A new list of series containing the data in the input series, but
        with their indices aligned. NaNs will be filled in for missing values.

    """
    head = many_series[0]
    tail = many_series[1:]
    n = len(head)
    if (isinstance(head, np.ndarray) and
            all(len(s) == n and isinstance(s, np.ndarray) for s in tail)):
        # optimization: ndarrays of the same length are already aligned
        return many_series

    # dataframe has no ``itervalues``
    return (
        v
        for _, v in iteritems(pd.concat(map(_to_pandas, many_series), axis=1))
    )


def alpha_beta(returns,
               factor_returns,
               risk_free=0.0,
               period=DAILY,
               annualization=None,
               out=None):
    """Calculates annualized alpha and beta.

    Parameters
    ----------
    returns : pd.Series
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series
         Daily noncumulative returns of the factor to which beta is
         computed. Usually a benchmark such as the market.
         - This is in the same style as returns.
    risk_free : int, float, optional
        Constant risk-free return throughout the period. For example, the
        interest rate on a three month us treasury bill.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    alpha : float
    beta : float
    """
    returns, factor_returns = _aligned_series(returns, factor_returns)

    return alpha_beta_aligned(
        returns,
        factor_returns,
        risk_free=risk_free,
        period=period,
        annualization=annualization,
        out=out,
    )


def roll_alpha_beta(returns, factor_returns, window=10, **kwargs):
    """
    Computes alpha and beta over a rolling window.

    Parameters
    ----------
    lhs : array-like
        The first array to pass to the rolling alpha-beta.
    rhs : array-like
        The second array to pass to the rolling alpha-beta.
    window : int
        Size of the rolling window in terms of the periodicity of the data.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.
    **kwargs
        Forwarded to :func:`~empyrical.alpha_beta`.
    """
    returns, factor_returns = _aligned_series(returns, factor_returns)

    return roll_alpha_beta_aligned(
        returns,
        factor_returns,
        window=window,
        **kwargs
    )


def alpha_beta_aligned(returns,
                       factor_returns,
                       risk_free=0.0,
                       period=DAILY,
                       annualization=None,
                       out=None):
    """Calculates annualized alpha and beta.

    If they are pd.Series, expects returns and factor_returns have already
    been aligned on their labels.  If np.ndarray, these arguments should have
    the same shape.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series or np.ndarray
         Daily noncumulative returns of the factor to which beta is
         computed. Usually a benchmark such as the market.
         - This is in the same style as returns.
    risk_free : int, float, optional
        Constant risk-free return throughout the period. For example, the
        interest rate on a three month us treasury bill.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    alpha : float
    beta : float
    """
    if out is None:
        out = np.empty(returns.shape[1:] + (2,), dtype='float64')

    b = beta_aligned(returns, factor_returns, risk_free, out=out[..., 1])
    alpha_aligned(
        returns,
        factor_returns,
        risk_free,
        period,
        annualization,
        out=out[..., 0],
        _beta=b,
    )

    return out


roll_alpha_beta_aligned = _create_binary_vectorized_roll_function(
    alpha_beta_aligned,
)


def alpha(returns,
          factor_returns,
          risk_free=0.0,
          period=DAILY,
          annualization=None,
          out=None,
          _beta=None):
    """Calculates annualized alpha.

    Parameters
    ----------
    returns : pd.Series
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series
        Daily noncumulative returns of the factor to which beta is
        computed. Usually a benchmark such as the market.
        - This is in the same style as returns.
    risk_free : int, float, optional
        Constant risk-free return throughout the period. For example, the
        interest rate on a three month us treasury bill.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.
        - See full explanation in :func:`~empyrical.stats.annual_return`.
    _beta : float, optional
        The beta for the given inputs, if already known. Will be calculated
        internally if not provided.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    float
        Alpha.
    """
    if not (isinstance(returns, np.ndarray) and
            isinstance(factor_returns, np.ndarray)):
        returns, factor_returns = _aligned_series(returns, factor_returns)

    return alpha_aligned(
        returns,
        factor_returns,
        risk_free=risk_free,
        period=period,
        annualization=annualization,
        out=out,
        _beta=_beta
    )


roll_alpha = _create_binary_vectorized_roll_function(alpha)


def alpha_aligned(returns,
                  factor_returns,
                  risk_free=0.0,
                  period=DAILY,
                  annualization=None,
                  out=None,
                  _beta=None):
    """Calculates annualized alpha.

    If they are pd.Series, expects returns and factor_returns have already
    been aligned on their labels.  If np.ndarray, these arguments should have
    the same shape.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series or np.ndarray
        Daily noncumulative returns of the factor to which beta is
        computed. Usually a benchmark such as the market.
        - This is in the same style as returns.
    risk_free : int, float, optional
        Constant risk-free return throughout the period. For example, the
        interest rate on a three month us treasury bill.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.
        - See full explanation in :func:`~empyrical.stats.annual_return`.
    _beta : float, optional
        The beta for the given inputs, if already known. Will be calculated
        internally if not provided.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    alpha : float
    """
    allocated_output = out is None
    if allocated_output:
        out = np.empty(returns.shape[1:], dtype='float64')

    if len(returns) < 2:
        out[()] = np.nan
        if returns.ndim == 1:
            out = out.item()
        return out

    ann_factor = annualization_factor(period, annualization)

    if _beta is None:
        _beta = beta_aligned(returns, factor_returns, risk_free)

    adj_returns = _adjust_returns(returns, risk_free)
    adj_factor_returns = _adjust_returns(factor_returns, risk_free)
    alpha_series = adj_returns - (_beta * adj_factor_returns)

    out = np.subtract(
        np.power(
            np.add(
                np.nanmean(alpha_series, axis=0, out=out),
                1,
                out=out
            ),
            ann_factor,
            out=out
        ),
        1,
        out=out
    )

    if allocated_output and isinstance(returns, pd.DataFrame):
        out = pd.Series(out)

    if returns.ndim == 1:
        out = out.item()

    return out


roll_alpha_aligned = _create_binary_vectorized_roll_function(alpha_aligned)


def beta(returns, factor_returns, risk_free=0.0, out=None):
    """Calculates beta.

    Parameters
    ----------
    returns : pd.Series
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series
         Daily noncumulative returns of the factor to which beta is
         computed. Usually a benchmark such as the market.
         - This is in the same style as returns.
    risk_free : int, float, optional
        Constant risk-free return throughout the period. For example, the
        interest rate on a three month us treasury bill.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    beta : float
    """
    if not (isinstance(returns, np.ndarray) and
            isinstance(factor_returns, np.ndarray)):
        returns, factor_returns = _aligned_series(returns, factor_returns)

    return beta_aligned(
        returns,
        factor_returns,
        risk_free=risk_free,
        out=out,
    )


roll_beta = _create_binary_vectorized_roll_function(beta)


def beta_aligned(returns, factor_returns, risk_free=0.0, out=None):
    """Calculates beta.

    If they are pd.Series, expects returns and factor_returns have already
    been aligned on their labels.  If np.ndarray, these arguments should have
    the same shape.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series or np.ndarray
         Daily noncumulative returns of the factor to which beta is
         computed. Usually a benchmark such as the market.
         - This is in the same style as returns.
    risk_free : int, float, optional
        Constant risk-free return throughout the period. For example, the
        interest rate on a three month us treasury bill.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    beta : float
        Beta.
    """
    # Cache these as locals since we're going to call them multiple times.
    nan = np.nan
    isnan = np.isnan

    returns_1d = returns.ndim == 1
    if returns_1d:
        returns = np.asanyarray(returns)[:, np.newaxis]

    if factor_returns.ndim == 1:
        factor_returns = np.asanyarray(factor_returns)[:, np.newaxis]

    N, M = returns.shape

    if out is None:
        out = np.full(M, nan)
    elif out.ndim == 0:
        out = out[np.newaxis]

    if len(returns) < 1 or len(factor_returns) < 2:
        out[()] = nan
        if returns_1d:
            out = out.item()
        return out

    # Copy N times as a column vector and fill with nans to have the same
    # missing value pattern as the dependent variable.
    #
    # PERF_TODO: We could probably avoid the space blowup by doing this in
    # Cython.

    # shape: (N, M)
    independent = np.where(
        isnan(returns),
        nan,
        factor_returns,
    )

    # Calculate beta as Cov(X, Y) / Cov(X, X).
    # https://en.wikipedia.org/wiki/Simple_linear_regression#Fitting_the_regression_line  # noqa
    #
    # NOTE: The usual formula for covariance is::
    #
    #    mean((X - mean(X)) * (Y - mean(Y)))
    #
    # However, we don't actually need to take the mean of both sides of the
    # product, because of the folllowing equivalence::
    #
    # Let X_res = (X - mean(X)).
    # We have:
    #
    #     mean(X_res * (Y - mean(Y))) = mean(X_res * (Y - mean(Y)))
    #                             (1) = mean((X_res * Y) - (X_res * mean(Y)))
    #                             (2) = mean(X_res * Y) - mean(X_res * mean(Y))
    #                             (3) = mean(X_res * Y) - mean(X_res) * mean(Y)
    #                             (4) = mean(X_res * Y) - 0 * mean(Y)
    #                             (5) = mean(X_res * Y)
    #
    #
    # The tricky step in the above derivation is step (4). We know that
    # mean(X_res) is zero because, for any X:
    #
    #     mean(X - mean(X)) = mean(X) - mean(X) = 0.
    #
    # The upshot of this is that we only have to center one of `independent`
    # and `dependent` when calculating covariances. Since we need the centered
    # `independent` to calculate its variance in the next step, we choose to
    # center `independent`.

    ind_residual = independent - np.nanmean(independent, axis=0)

    covariances = np.nanmean(ind_residual * returns, axis=0)

    # We end up with different variances in each column here because each
    # column may have a different subset of the data dropped due to missing
    # data in the corresponding dependent column.
    # shape: (M,)
    np.square(ind_residual, out=ind_residual)
    independent_variances = np.nanmean(ind_residual, axis=0)
    independent_variances[independent_variances < 1.0e-30] = np.nan

    np.divide(covariances, independent_variances, out=out)

    if returns_1d:
        out = out.item()

    return out


roll_beta_aligned = _create_binary_vectorized_roll_function(beta_aligned)


def stability_of_timeseries(returns):
    """Determines R-squared of a linear fit to the cumulative
    log returns. Computes an ordinary least squares linear fit,
    and returns R-squared.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.

    Returns
    -------
    float
        R-squared.

    """
    if len(returns) < 2:
        return np.nan

    returns = np.asanyarray(returns)
    returns = returns[~np.isnan(returns)]

    cum_log_returns = np.log1p(returns).cumsum()
    rhat = stats.linregress(np.arange(len(cum_log_returns)),
                            cum_log_returns)[2]

    return rhat ** 2


def tail_ratio(returns):
    """Determines the ratio between the right (95%) and left tail (5%).

    For example, a ratio of 0.25 means that losses are four times
    as bad as profits.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
         - See full explanation in :func:`~empyrical.stats.cum_returns`.

    Returns
    -------
    tail_ratio : float
    """

    if len(returns) < 1:
        return np.nan

    returns = np.asanyarray(returns)
    # Be tolerant of nan's
    returns = returns[~np.isnan(returns)]
    if len(returns) < 1:
        return np.nan

    return np.abs(np.percentile(returns, 95)) / \
        np.abs(np.percentile(returns, 5))


def capture(returns, factor_returns, period=DAILY):
    """Compute capture ratio.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series or np.ndarray
        Noncumulative returns of the factor to which beta is
        computed. Usually a benchmark such as the market.
        - This is in the same style as returns.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    Returns
    -------
    capture_ratio : float

    Note
    ----
    See http://www.investopedia.com/terms/u/up-market-capture-ratio.asp for
    details.
    """
    return (annual_return(returns, period=period) /
            annual_return(factor_returns, period=period))


def beta_fragility_heuristic(returns, factor_returns):
    """Estimate fragility to drops in beta.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series or np.ndarray
         Daily noncumulative returns of the factor to which beta is
         computed. Usually a benchmark such as the market.
         - This is in the same style as returns.

    Returns
    -------
    float, np.nan
        The beta fragility of the strategy.

    Note
    ----
    A negative return value indicates potential losses
    could follow volatility in beta.
    The magnitude of the negative value indicates the size of
    the potential loss.
    seealso::
    `A New Heuristic Measure of Fragility and
Tail Risks: Application to Stress Testing`
        https://www.imf.org/external/pubs/ft/wp/2012/wp12216.pdf
        An IMF Working Paper describing the heuristic
    """
    if len(returns) < 3 or len(factor_returns) < 3:
        return np.nan

    return beta_fragility_heuristic_aligned(
        *_aligned_series(returns, factor_returns))


def beta_fragility_heuristic_aligned(returns, factor_returns):
    """Estimate fragility to drops in beta

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series or np.ndarray
         Daily noncumulative returns of the factor to which beta is
         computed. Usually a benchmark such as the market.
         - This is in the same style as returns.

    Returns
    -------
    float, np.nan
        The beta fragility of the strategy.

    Note
    ----
    If they are pd.Series, expects returns and factor_returns have already
    been aligned on their labels.  If np.ndarray, these arguments should have
    the same shape.
    seealso::
    `A New Heuristic Measure of Fragility and
Tail Risks: Application to Stress Testing`
        https://www.imf.org/external/pubs/ft/wp/2012/wp12216.pdf
        An IMF Working Paper describing the heuristic
    """
    if len(returns) < 3 or len(factor_returns) < 3:
        return np.nan

    # combine returns and factor returns into pairs
    returns_series = pd.Series(returns)
    factor_returns_series = pd.Series(factor_returns)
    pairs = pd.concat([returns_series, factor_returns_series], axis=1)
    pairs.columns = ['returns', 'factor_returns']

    # exclude any rows where returns are nan
    pairs = pairs.dropna()
    # sort by beta
    pairs = pairs.sort_values(by='factor_returns')

    # find the three vectors, using median of 3
    start_index = 0
    mid_index = int(np.around(len(pairs) / 2, 0))
    end_index = len(pairs) - 1

    (start_returns, start_factor_returns) = pairs.iloc[start_index]
    (mid_returns, mid_factor_returns) = pairs.iloc[mid_index]
    (end_returns, end_factor_returns) = pairs.iloc[end_index]

    factor_returns_range = (end_factor_returns - start_factor_returns)
    start_returns_weight = 0.5
    end_returns_weight = 0.5

    # find weights for the start and end returns
    # using a convex combination
    if not factor_returns_range == 0:
        start_returns_weight = \
            (mid_factor_returns - start_factor_returns) / \
            factor_returns_range
        end_returns_weight = \
            (end_factor_returns - mid_factor_returns) / \
            factor_returns_range

    # calculate fragility heuristic
    heuristic = (start_returns_weight * start_returns) + \
                (end_returns_weight * end_returns) - mid_returns

    return heuristic


def gpd_risk_estimates(returns, var_p=0.01):
    """Estimate VaR and ES using the Generalized Pareto Distribution (GPD)

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    var_p : float
        The percentile to use for estimating the VaR and ES

    Returns
    -------
    [threshold, scale_param, shape_param, var_estimate, es_estimate]
        : list[float]
        threshold - the threshold use to cut off exception tail losses
        scale_param - a parameter (often denoted by sigma, capturing the
            scale, related to variance)
        shape_param - a parameter (often denoted by xi, capturing the shape or
            type of the distribution)
        var_estimate - an estimate for the VaR for the given percentile
        es_estimate - an estimate for the ES for the given percentile

    Note
    ----
    seealso::
    `An Application of Extreme Value Theory for
Measuring Risk <https://link.springer.com/article/10.1007/s10614-006-9025-7>`
        A paper describing how to use the Generalized Pareto
        Distribution to estimate VaR and ES.
    """
    if len(returns) < 3:
        result = np.zeros(5)
        if isinstance(returns, pd.Series):
            result = pd.Series(result)
        return result
    return gpd_risk_estimates_aligned(*_aligned_series(returns, var_p))


def gpd_risk_estimates_aligned(returns, var_p=0.01):
    """Estimate VaR and ES using the Generalized Pareto Distribution (GPD)

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    var_p : float
        The percentile to use for estimating the VaR and ES

    Returns
    -------
    [threshold, scale_param, shape_param, var_estimate, es_estimate]
        : list[float]
        threshold - the threshold use to cut off exception tail losses
        scale_param - a parameter (often denoted by sigma, capturing the
            scale, related to variance)
        shape_param - a parameter (often denoted by xi, capturing the shape or
            type of the distribution)
        var_estimate - an estimate for the VaR for the given percentile
        es_estimate - an estimate for the ES for the given percentile

    Note
    ----
    seealso::
    `An Application of Extreme Value Theory for
Measuring Risk <https://link.springer.com/article/10.1007/s10614-006-9025-7>`
        A paper describing how to use the Generalized Pareto
        Distribution to estimate VaR and ES.
    """
    result = np.zeros(5)
    if not len(returns) < 3:

        DEFAULT_THRESHOLD = 0.2
        MINIMUM_THRESHOLD = 0.000000001

        try:
            returns_array = pd.Series(returns).to_numpy()
        except AttributeError:
            # while zipline requires support for pandas < 0.25
            returns_array = pd.Series(returns).as_matrix()

        flipped_returns = -1 * returns_array
        losses = flipped_returns[flipped_returns > 0]
        threshold = DEFAULT_THRESHOLD
        finished = False
        scale_param = 0
        shape_param = 0
        while not finished and threshold > MINIMUM_THRESHOLD:
            losses_beyond_threshold = \
                losses[losses >= threshold]
            param_result = \
                gpd_loglikelihood_minimizer_aligned(losses_beyond_threshold)
            if (param_result[0] is not False and
                    param_result[1] is not False):
                scale_param = param_result[0]
                shape_param = param_result[1]
                var_estimate = gpd_var_calculator(threshold, scale_param,
                                                  shape_param, var_p,
                                                  len(losses),
                                                  len(losses_beyond_threshold))
                # non-negative shape parameter is required for fat tails
                # non-negative VaR estimate is required for loss of some kind
                if (shape_param > 0 and var_estimate > 0):
                    finished = True
            if (not finished):
                threshold = threshold / 2
        if (finished):
            es_estimate = gpd_es_calculator(var_estimate, threshold,
                                            scale_param, shape_param)
            result = np.array([threshold, scale_param, shape_param,
                               var_estimate, es_estimate])
    if isinstance(returns, pd.Series):
        result = pd.Series(result)
    return result


def gpd_es_calculator(var_estimate, threshold, scale_param,
                      shape_param):
    result = 0
    if ((1 - shape_param) != 0):
        # this formula is from Gilli and Kellezi pg. 8
        var_ratio = (var_estimate / (1 - shape_param))
        param_ratio = ((scale_param - (shape_param * threshold)) /
                       (1 - shape_param))
        result = var_ratio + param_ratio
    return result


def gpd_var_calculator(threshold, scale_param, shape_param,
                       probability, total_n, exceedance_n):
    result = 0
    if (exceedance_n > 0 and shape_param > 0):
        # this formula is from Gilli and Kellezi pg. 12
        param_ratio = scale_param / shape_param
        prob_ratio = (total_n / exceedance_n) * probability
        result = threshold + (param_ratio *
                              (pow(prob_ratio, -shape_param) - 1))
    return result


def gpd_loglikelihood_minimizer_aligned(price_data):
    result = [False, False]
    DEFAULT_SCALE_PARAM = 1
    DEFAULT_SHAPE_PARAM = 1
    if (len(price_data) > 0):
        gpd_loglikelihood_lambda = \
            gpd_loglikelihood_factory(price_data)
        optimization_results = \
            optimize.minimize(gpd_loglikelihood_lambda,
                              [DEFAULT_SCALE_PARAM,
                               DEFAULT_SHAPE_PARAM],
                              method='Nelder-Mead')
        if optimization_results.success:
            resulting_params = optimization_results.x
            if len(resulting_params) == 2:
                result[0] = resulting_params[0]
                result[1] = resulting_params[1]
    return result


def gpd_loglikelihood_factory(price_data):
    return lambda params: gpd_loglikelihood(params, price_data)


def gpd_loglikelihood(params, price_data):
    if (params[1] != 0):
        return -gpd_loglikelihood_scale_and_shape(params[0],
                                                  params[1],
                                                  price_data)
    else:
        return -gpd_loglikelihood_scale_only(params[0], price_data)


def gpd_loglikelihood_scale_and_shape_factory(price_data):
    # minimize a function of two variables requires a list of params
    # we are expecting the lambda below to be called as follows:
    # parameters = [scale, shape]
    # the final outer negative is added because scipy only minimizes
    return lambda params: \
        -gpd_loglikelihood_scale_and_shape(params[0],
                                           params[1],
                                           price_data)


def gpd_loglikelihood_scale_and_shape(scale, shape, price_data):
    n = len(price_data)
    result = -1 * float_info.max
    if (scale != 0):
        param_factor = shape / scale
        if (shape != 0 and param_factor >= 0 and scale >= 0):
            result = ((-n * np.log(scale)) -
                      (((1 / shape) + 1) *
                       (np.log((shape / scale * price_data) + 1)).sum()))
    return result


def gpd_loglikelihood_scale_only_factory(price_data):
    # the negative is added because scipy only minimizes
    return lambda scale: \
        -gpd_loglikelihood_scale_only(scale, price_data)


def gpd_loglikelihood_scale_only(scale, price_data):
    n = len(price_data)
    data_sum = price_data.sum()
    result = -1 * float_info.max
    if (scale >= 0):
        result = ((-n * np.log(scale)) - (data_sum / scale))
    return result


def up_capture(returns, factor_returns, **kwargs):
    """
    Compute the capture ratio for periods when the benchmark return is positive

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series or np.ndarray
        Noncumulative returns of the factor to which beta is
        computed. Usually a benchmark such as the market.
        - This is in the same style as returns.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    Returns
    -------
    up_capture : float

    Note
    ----
    See http://www.investopedia.com/terms/u/up-market-capture-ratio.asp for
    more information.
    """
    return up(returns, factor_returns, function=capture, **kwargs)


def down_capture(returns, factor_returns, **kwargs):
    """
    Compute the capture ratio for periods when the benchmark return is negative

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series or np.ndarray
        Noncumulative returns of the factor to which beta is
        computed. Usually a benchmark such as the market.
        - This is in the same style as returns.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    Returns
    -------
    down_capture : float

    Note
    ----
    See http://www.investopedia.com/terms/d/down-market-capture-ratio.asp for
    more information.
    """
    return down(returns, factor_returns, function=capture, **kwargs)


def up_down_capture(returns, factor_returns, **kwargs):
    """
    Computes the ratio of up_capture to down_capture.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series or np.ndarray
        Noncumulative returns of the factor to which beta is
        computed. Usually a benchmark such as the market.
        - This is in the same style as returns.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    Returns
    -------
    up_down_capture : float
        the updown capture ratio
    """
    return (up_capture(returns, factor_returns, **kwargs) /
            down_capture(returns, factor_returns, **kwargs))


def up_alpha_beta(returns, factor_returns, **kwargs):
    """
    Computes alpha and beta for periods when the benchmark return is positive.

    Parameters
    ----------
    see documentation for `alpha_beta`.

    Returns
    -------
    float
        Alpha.
    float
        Beta.
    """
    return up(returns, factor_returns, function=alpha_beta_aligned, **kwargs)


def down_alpha_beta(returns, factor_returns, **kwargs):
    """
    Computes alpha and beta for periods when the benchmark return is negative.

    Parameters
    ----------
    see documentation for `alpha_beta`.

    Returns
    -------
    alpha : float
    beta : float
    """
    return down(returns, factor_returns, function=alpha_beta_aligned, **kwargs)


def roll_up_capture(returns, factor_returns, window=10, **kwargs):
    """
    Computes the up capture measure over a rolling window.
    see documentation for :func:`~empyrical.stats.up_capture`.
    (pass all args, kwargs required)

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.

    factor_returns : pd.Series or np.ndarray
        Noncumulative returns of the factor to which beta is
        computed. Usually a benchmark such as the market.
        - This is in the same style as returns.

    window : int, required
        Size of the rolling window in terms of the periodicity of the data.
        - eg window = 60, periodicity=DAILY, represents a rolling 60 day window
    """
    return roll(returns, factor_returns, window=window, function=up_capture,
                **kwargs)


def roll_down_capture(returns, factor_returns, window=10, **kwargs):
    """
    Computes the down capture measure over a rolling window.
    see documentation for :func:`~empyrical.stats.down_capture`.
    (pass all args, kwargs required)

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.

    factor_returns : pd.Series or np.ndarray
        Noncumulative returns of the factor to which beta is
        computed. Usually a benchmark such as the market.
        - This is in the same style as returns.

    window : int, required
        Size of the rolling window in terms of the periodicity of the data.
        - eg window = 60, periodicity=DAILY, represents a rolling 60 day window
    """
    return roll(returns, factor_returns, window=window, function=down_capture,
                **kwargs)


def roll_up_down_capture(returns, factor_returns, window=10, **kwargs):
    """
    Computes the up/down capture measure over a rolling window.
    see documentation for :func:`~empyrical.stats.up_down_capture`.
    (pass all args, kwargs required)

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.

    factor_returns : pd.Series or np.ndarray
        Noncumulative returns of the factor to which beta is
        computed. Usually a benchmark such as the market.
        - This is in the same style as returns.

    window : int, required
        Size of the rolling window in terms of the periodicity of the data.
        - eg window = 60, periodicity=DAILY, represents a rolling 60 day window
    """
    return roll(returns, factor_returns, window=window,
                function=up_down_capture, **kwargs)


def value_at_risk(returns, cutoff=0.05):
    """
    Value at risk (VaR) of a returns stream.

    Parameters
    ----------
    returns : pandas.Series or 1-D numpy.array
        Non-cumulative daily returns.
    cutoff : float, optional
        Decimal representing the percentage cutoff for the bottom percentile of
        returns. Defaults to 0.05.

    Returns
    -------
    VaR : float
        The VaR value.
    """
    return np.percentile(returns, 100 * cutoff)


def conditional_value_at_risk(returns, cutoff=0.05):
    """
    Conditional value at risk (CVaR) of a returns stream.

    CVaR measures the expected single-day returns of an asset on that asset's
    worst performing days, where "worst-performing" is defined as falling below
    ``cutoff`` as a percentile of all daily returns.

    Parameters
    ----------
    returns : pandas.Series or 1-D numpy.array
        Non-cumulative daily returns.
    cutoff : float, optional
        Decimal representing the percentage cutoff for the bottom percentile of
        returns. Defaults to 0.05.

    Returns
    -------
    CVaR : float
        The CVaR value.
    """
    # PERF: Instead of using the 'value_at_risk' function to find the cutoff
    # value, which requires a call to numpy.percentile, determine the cutoff
    # index manually and partition out the lowest returns values. The value at
    # the cutoff index should be included in the partition.
    cutoff_index = int((len(returns) - 1) * cutoff)
    return np.mean(np.partition(returns, cutoff_index)[:cutoff_index + 1])


SIMPLE_STAT_FUNCS = [
    cum_returns_final,
    annual_return,
    annual_volatility,
    sharpe_ratio,
    calmar_ratio,
    stability_of_timeseries,
    max_drawdown,
    omega_ratio,
    sortino_ratio,
    stats.skew,
    stats.kurtosis,
    tail_ratio,
    cagr,
    value_at_risk,
    conditional_value_at_risk,
]

FACTOR_STAT_FUNCS = [
    excess_sharpe,
    alpha,
    beta,
    beta_fragility_heuristic,
    gpd_risk_estimates,
    capture,
    up_capture,
    down_capture
]



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License. 
Created on 2025/6/27 18:51
Email: yundi.xxii@outlook.com
Description: 
---------------------------------------------
"""

# quda/ml/trainer/pytorch.py

from .base import BaseIncrementalTrainer
import pytorch_lightning as pl
import torch
from mlflow.models.signature import infer_signature


class PyTorchLightningTrainer(BaseIncrementalTrainer):
    def __init__(self, model: pl.LightningModule, datamodule: pl.LightningDataModule, experiment_name="PTL-Incremental"):
        super().__init__(experiment_name=experiment_name)
        self.model = model
        self.datamodule = datamodule
        self.trainer = None

    def create_dataset(self, data, features: list[str], target: str):
        # 将 Polars DataFrame 转换为 Dataset / DataLoader
        return self.datamodule(data, features, target)

    def fit(self, train_data, valid_data=None, init_model=None, evals_result: dict | None = None, **model_params):
        trainer_kwargs = {
            "max_epochs": model_params.get("max_epochs", 10),
            "callbacks": [self._mlflow_callback(evals_result)],
            "logger": False,
        }
        self.trainer = pl.Trainer(**trainer_kwargs)
        self.trainer.fit(self.model, datamodule=self.datamodule)
        return self.model

    def _mlflow_callback(self, evals_result):
        """MLflow logging callback"""
        class MLFlowLoggerCallback(pl.Callback):
            def on_validation_end(self, trainer, pl_module):
                metrics = trainer.callback_metrics
                for key, value in metrics.items():
                    mlflow.log_metric(key, value.item(), step=trainer.global_step)
                evals_result.update(metrics)
        return MLFlowLoggerCallback()



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License. 
Created on 2025/7/9 15:40
Email: yundi.xxii@outlook.com
Description: tick数据清洗
---------------------------------------------
"""
import xcals

import quda
import polars as pl
from quda.factor import INDEX, TIME_FORMAT
from quda.data.tables import TB_STOCK_TICK
from pandas import Timedelta
import ygo
import ylog


DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
END_CALL_AUC = "09:25:00"
END_AM = "11:30:00"
END_PM = "15:00:00"

BEG_AM = "09:30:00"
BEG_PM = "13:00:00"
END_TICK = "16:00:00"


def read_tick(date: str, beg_time: str = None, end_time: str = None, ) -> pl.LazyFrame:
    beg_time = beg_time if beg_time else "09:14:00"
    end_time = end_time if end_time else END_TICK
    sql = f"""
    SELECT *
    FROM {TB_STOCK_TICK}
    WHERE date = '{date}'
        AND time BETWEEN '{beg_time}' AND '{end_time}';
    """
    df = quda.sql(sql, lazy=True).with_columns(pl.col("close").alias("price")).drop("close")
    return df

def _to_kline(date, beg_time: str = "09:23:00", end_time: str = "16:00:00", freq: str = "1min") -> pl.LazyFrame:
    """合成分钟kline"""
    beg_time = beg_time if beg_time is not None else "09:23:00"
    end_time = end_time if end_time is not None else END_TICK
    df = read_tick(date, beg_time, end_time).select("date", "time", "asset", "high", "low", "price", "volume",
                                                    "amount", "prev_close", "limit_up", "limit_down")
    secs = Timedelta(freq).seconds
    over_spec = {"partition_by": "asset", "order_by": "time"}
    df = (
        df
        .with_columns(pl.lit(date).alias("date"),
                      _time_=pl.col("time").str.to_datetime(TIME_FORMAT).dt.truncate(f"{secs}s"), )
        .with_columns(diff_v=pl.col("volume").diff().over(**over_spec),
                      diff_p=pl.col("price").diff().over(**over_spec))
        .with_columns(time_str=pl.col("_time_").dt.time().cast(pl.Utf8), )
        .with_columns(pl.when(pl.col("time_str").is_in([END_CALL_AUC, END_AM, END_PM]))
                      .then(pl.col("time_str"))
                      .otherwise((pl.col("_time_") + pl.duration(seconds=secs)).dt.time().cast(pl.Utf8))
                      .alias("time"))
    )
    # 合成分钟kline: 最高价、最低价、开盘价、收盘价、成交量、成交额
    df = df.select("date", "time", "asset", "prev_close", "limit_up", "limit_down", "price", "volume", "amount",
                   pl.col("high").alias("cum_h"),
                   pl.col("low").alias("cum_l"),
                   pl.col("diff_v"),
                   pl.col("diff_p"))
    kline = (
        df
        .with_columns(price=pl.when(pl.col("price") > 0).then(pl.col("price")).otherwise(None))
        .with_columns(
            price=pl.when((pl.col("diff_v") > 0) | (pl.col("diff_p").abs() > 0)).then(pl.col("price")).otherwise(None))
        .group_by(INDEX)
        .agg(prev_close=pl.col("prev_close").max(),
             limit_up=pl.col("limit_up").max(),
             limit_down=pl.col("limit_down").max(),
             open=pl.col("price").drop_nulls().first(),
             high=pl.col("cum_h").last(),
             low=pl.col("cum_l").last(),
             close=pl.col("price").drop_nulls().last(),
             cum_vol=pl.col("volume").last(),
             cum_amt=pl.col("amount").last(),
             max_p=pl.col("price").max(),
             min_p=pl.col("price").min(), )
        .sort(INDEX)
        .with_columns(diff_h=pl.col("high").diff().over(**over_spec),
                      diff_l=pl.col("low").diff().over(**over_spec),
                      )
        .with_columns(high=pl.when(pl.col("diff_h") > 0).then(pl.col("high")).otherwise("max_p"),
                      low=pl.when(pl.col("diff_l") < 0).then(pl.col("low")).otherwise("min_p"), )
        .drop("max_p", "min_p", "diff_h", "diff_l", )
    )
    # if beg_time <= END_CALL_AUC <= end_time:
    if end_time < BEG_AM:
        kline = (
            kline
            .with_columns(*[
                pl.when(pl.col("time") == END_CALL_AUC).then(pl.col(field).fill_null(pl.col("prev_close"))).otherwise(
                    pl.col(field)) for field in ["open", "high", "low", "close"]])
            .with_columns(
                *[pl.col(field).forward_fill().over(**over_spec) for field in ["open", "high", "low", "close"]])
        )
    return kline

def agg_kline(kline: pl.LazyFrame, kline_time: str | None = None) -> pl.LazyFrame:
    """kline聚合：由多个K线聚合成一条kline"""
    kline = (
        kline
        .sort(INDEX)
        .group_by("date", "asset")
        .agg(time=pl.col("time").last(),
             prev_close=pl.col("prev_close").max(),
             limit_up=pl.col("limit_up").max(),
             limit_down=pl.col("limit_down").max(),
             open=pl.col("open").drop_nulls().first(),
             high=pl.col("high").max(),
             low=pl.col("low").min(),
             close=pl.col("close").drop_nulls().last(),
             cum_vol=pl.col("cum_vol").last(),
             cum_amt=pl.col("cum_amt").last())
    )
    if kline_time is not None:
        kline = (
            kline
            .with_columns(time=pl.lit(kline_time))
        )
    return kline.select(*INDEX, "prev_close", "limit_up", "limit_down", "open", "high", "low", "close", "cum_vol", "cum_amt")

def to_kline_oneday(date, freq: str = "1min"):
    """一次性合成一天的所有kline"""
    kline_all = _to_kline(date, beg_time="09:23:00", end_time="16:00:00", freq=freq)
    # 集合竞价
    kline_beg = kline_all.filter(pl.col("time") < BEG_AM)
    kline_beg = agg_kline(kline_beg, kline_time=BEG_AM)
    # 上午
    kline_am = kline_all.filter(pl.col("time") > BEG_AM, pl.col("time") < END_AM)
    # 中午盘
    kline_skip = kline_all.filter(pl.col("time") >= END_AM, pl.col("time") <= BEG_PM)
    kline_skip = agg_kline(kline_skip, kline_time=END_AM)
    # 下午
    kline_pm = kline_all.filter(pl.col("time") > BEG_PM, pl.col("time") < END_PM)
    # 尾盘
    kline_end = kline_all.filter(pl.col("time") >= END_PM)
    # 收盘后
    kline_end = agg_kline(kline_end, kline_time=END_PM)
    klines = [kline_beg, kline_am, kline_skip, kline_pm, kline_end, ]
    kline = pl.concat(klines, how="vertical").sort(INDEX)
    # 对齐所有symbol(由于某些symbol不活跃而导致某些时段没有tick数据，需要修复)
    full_index_t = kline.select("date", "time").collect().unique()
    full_index_a = kline.select("asset").collect().unique()
    full_index = full_index_t.join(full_index_a, how="cross").lazy()
    over_spec = {"partition_by": "asset", "order_by": "time"}
    kline = (
        full_index
        .join(kline, on=INDEX, how="left")
        .sort(INDEX)
        .with_columns(pl.all().exclude("open", "high", "low").forward_fill().over(**over_spec))
    )
    # 调整 volume / amount / 以及其他一些价格字段
    kline = (
        kline
        .with_columns(volume=pl.col("cum_vol").diff().over(**over_spec).fill_null(pl.col("cum_vol")),
                      amount=pl.col("cum_amt").diff().over(**over_spec).fill_null(pl.col("cum_amt")),
                      last_close=pl.col("close").shift(1).forward_fill().over(**over_spec))
        .with_columns(
            *[pl.when(pl.col("volume") == 0).then(pl.col("close").fill_null(pl.col("last_close"))).otherwise(
                pl.col(field)).alias(field) for
              field in ["open", "high", "low", "close"]])
        .with_columns(
            *[pl.col(field).fill_null(pl.col("last_close")).alias(field) for field in ["open", "high", "low", "close"]])
    )
    kline = kline.sort(INDEX).select(*INDEX, "prev_close", "limit_up", "limit_down", "open", "high", "low", "close",
                                     "cum_vol", "cum_amt", "volume", "amount", )
    kline = kline.with_columns(pl.col(pl.Decimal).cast(pl.Decimal(16, 4)))
    return kline.drop("cum_vol", "cum_amt").collect()

def _to_tick(date: str, beg_time: str, end_time: str, freq: str) -> pl.LazyFrame:
    tick = read_tick(date, beg_time, end_time).drop("high", "low", "price", "volume", "amount")
    secs = Timedelta(freq).seconds
    tick = (
        tick
        .with_columns(pl.lit(date).alias("date"),
                      _time_=pl.col("time").str.to_datetime(TIME_FORMAT).dt.truncate(f"{secs}s"), )
        .with_columns(time_str=pl.col("_time_").dt.time().cast(pl.Utf8), )
        .with_columns(pl.when(pl.col("time_str").is_in([END_CALL_AUC, END_AM, END_PM]))
                      .then(pl.col("time_str"))
                      .otherwise((pl.col("_time_") + pl.duration(seconds=secs)).dt.time().cast(pl.Utf8))
                      .alias("time"))
        .select("date", "time", "asset",
                "ask_p1", "ask_p2", "ask_p3", "ask_p4", "ask_p5",
                "ask_v1", "ask_v2", "ask_v3", "ask_v4", "ask_v5",
                "bid_p1", "bid_p2", "bid_p3", "bid_p4", "bid_p5",
                "bid_v1", "bid_v2", "bid_v3", "bid_v4", "bid_v5")
        .sort(INDEX)
    )
    return (
        tick
        .group_by(INDEX)
        .agg(pl.all().last().cast(pl.Decimal(16, 4)))
    )


def agg_tick(ticks: pl.DataFrame, tick_time: str) -> pl.LazyFrame:
    """tick聚合：由多个tick聚合成一个tick"""
    fields = ["time", "ask_p1", "ask_p2", "ask_p3", "ask_p4", "ask_p5",
              "ask_v1", "ask_v2", "ask_v3", "ask_v4", "ask_v5",
              "bid_p1", "bid_p2", "bid_p3", "bid_p4", "bid_p5",
              "bid_v1", "bid_v2", "bid_v3", "bid_v4", "bid_v5"]
    tick = (
        ticks
        .sort(INDEX)
        .group_by("date", "asset")
        .agg(*[pl.col(field).last() for field in fields])
    )
    if tick_time is not None:
        tick = (
            tick
            .with_columns(time=pl.lit(tick_time))
        )
    return tick.select(*INDEX, *fields[1:])

def to_tick_oneday(date, freq: str = "1min"):
    """一次性合成一天的所有tick"""
    tick_all = _to_tick(date, beg_time="09:23:00", end_time="16:00:00", freq=freq)
    # 集合竞价
    tick_beg = tick_all.filter(pl.col("time") < BEG_AM)
    tick_beg = agg_tick(tick_beg, tick_time=BEG_AM)
    # 上午
    tick_am = tick_all.filter(pl.col("time") > BEG_AM, pl.col("time") < END_AM)
    # 中午盘
    tick_skip = tick_all.filter(pl.col("time") >= END_AM, pl.col("time") <= BEG_PM)
    tick_skip = agg_tick(tick_skip, tick_time=END_AM)
    # 下午
    tick_pm = tick_all.filter(pl.col("time") > BEG_PM, pl.col("time") < END_PM)
    # 尾盘
    tick_end = tick_all.filter(pl.col("time") >= END_PM)
    # 收盘后
    tick_end = agg_tick(tick_end, tick_time=END_PM)
    ticks = [tick_beg, tick_am, tick_skip, tick_pm, tick_end]
    tick = pl.concat(ticks, how="vertical").sort(INDEX)
    # 对齐所有symbol(由于某些symbol不活跃而导致某些时段没有tick数据，需要修复)
    full_index_t = tick.select("date", "time").collect().unique()
    full_index_a = tick.select("asset").collect().unique()
    full_index = full_index_t.join(full_index_a, how="cross").lazy()
    over_spec = {"partition_by": "asset", "order_by": "time"}
    return (
        full_index
        .join(tick, on=INDEX, how="left")
        .sort(INDEX)
        .with_columns(pl.all().forward_fill().over(**over_spec))

    )

def save_ytick_oneday(date: str, tb_name: str):
    """tick数据清洗, 时间对齐, ohlc 修复"""
    try:
        cnt = quda.sql(f"select count(distinct date) from {tb_name} where date='{date}';").collect().item()
        if cnt > 0:
            return
    except Exception as e:
        ylog.error(e)
        pass
    # 合成3s kline
    kline_3s = to_kline_oneday(date, freq="3s")
    tick_3s = to_tick_oneday(date, freq="3s")
    ytick = (
        kline_3s
        .join(tick_3s, on=INDEX, how="left")
        .sort(INDEX)
    )
    quda.put(ytick, tb_name=tb_name, partitions=["date"], )
    return


def save_ytick(beg_date: str, end_date: str, tb_name: str, n_jobs: int = 5):
    """保存ytick数据"""
    trade_days = xcals.get_tradingdays(beg_date, end_date)
    with ygo.pool(n_jobs=n_jobs, backend="threading") as go:
        for date in trade_days:
            go.submit(save_ytick_oneday, job_name="Saving ytick")(tb_name=tb_name, date=date)
        go.do()


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import math
import importlib
from typing import Iterable

import plotly.offline as py
import plotly.graph_objs as go

from plotly.subplots import make_subplots
from plotly.figure_factory import create_distplot
from IPython import get_ipython


class BaseGraph:
    _name = None

    def __init__(
        self, df=None, layout: dict = None, graph_kwargs: dict = None, name_dict: dict = None, **kwargs
    ):
        """

        :param df:
        :param layout:
        :param graph_kwargs:
        :param name_dict:
        :param kwargs:
            layout: dict
                go.Layout parameters
            graph_kwargs: dict
                Graph parameters, eg: go.Bar(**graph_kwargs)
        """
        self._df = df

        self._layout = dict() if layout is None else layout
        self._graph_kwargs = dict() if graph_kwargs is None else graph_kwargs
        self._name_dict = name_dict

        self.data = None

        self._init_parameters(**kwargs)
        self._init_data()

    def _init_data(self):
        """

        :return:
        """
        if self._df.empty:
            raise ValueError("df is empty.")

        self.data = self._get_data()

    def _init_parameters(self, **kwargs):
        """

        :param kwargs
        """

        # Instantiate graphics parameters
        self._graph_type = self._name.lower().capitalize()

        # Displayed column name
        if self._name_dict is None:
            self._name_dict = {_item: _item for _item in self._df.columns}

    @staticmethod
    def get_instance_with_graph_parameters(graph_type: str = None, **kwargs):
        """

        :param graph_type:
        :param kwargs:
        :return:
        """
        try:
            _graph_module = importlib.import_module("plotly.graph_objs")
            _graph_class = getattr(_graph_module, graph_type)
        except AttributeError:
            _graph_module = importlib.import_module("alphalens.report_qlib.graph")
            _graph_class = getattr(_graph_module, graph_type)
        return _graph_class(**kwargs)

    @staticmethod
    def show_graph_in_notebook(figure_list: Iterable[go.Figure] = None):
        """

        :param figure_list:
        :return:
        """
        py.init_notebook_mode()
        for _fig in figure_list:
            # NOTE: displays figures: https://plotly.com/python/renderers/
            # default: plotly_mimetype+notebook
            # support renderers: import plotly.io as pio; print(pio.renderers)
            renderer = None
            try:
                # in notebook
                _ipykernel = str(type(get_ipython()))
                if "google.colab" in _ipykernel:
                    renderer = "colab"
            except NameError:
                pass

            _fig.show(renderer=renderer)

    def _get_layout(self) -> go.Layout:
        """

        :return:
        """
        return go.Layout(**self._layout)

    def _get_data(self) -> list:
        """

        :return:
        """

        _data = [
            self.get_instance_with_graph_parameters(
                graph_type=self._graph_type, x=self._df.index, y=self._df[_col], name=_name, **self._graph_kwargs
            )
            for _col, _name in self._name_dict.items()
        ]
        return _data

    @property
    def figure(self) -> go.Figure:
        """

        :return:
        """
        _figure = go.Figure(data=self.data, layout=self._get_layout())
        # NOTE: Use the default theme from plotly version 3.x, template=None
        _figure["layout"].update(template=None)
        return _figure


class ScatterGraph(BaseGraph):
    _name = "scatter"


class BarGraph(BaseGraph):
    _name = "bar"


class DistplotGraph(BaseGraph):
    _name = "distplot"

    def _get_data(self):
        """

        :return:
        """
        _t_df = self._df.dropna()
        _data_list = [_t_df[_col] for _col in self._name_dict]
        _label_list = list(self._name_dict.values())
        _fig = create_distplot(_data_list, _label_list, show_rug=False, **self._graph_kwargs)

        return _fig["data"]


class HeatmapGraph(BaseGraph):
    _name = "heatmap"

    def _get_data(self):
        """

        :return:
        """
        _data = [
            self.get_instance_with_graph_parameters(
                graph_type=self._graph_type,
                x=self._df.columns,
                y=self._df.index,
                z=self._df.values.tolist(),
                **self._graph_kwargs,
            )
        ]
        return _data


class HistogramGraph(BaseGraph):
    _name = "histogram"

    def _get_data(self):
        """

        :return:
        """
        _data = [
            self.get_instance_with_graph_parameters(
                graph_type=self._graph_type, x=self._df[_col], name=_name, **self._graph_kwargs
            )
            for _col, _name in self._name_dict.items()
        ]
        return _data


class SubplotsGraph:
    """Create subplots same as df.plot(subplots=True)

    Simple package for `plotly.tools.subplots`
    """

    def __init__(
        self,
        df=None,
        kind_map: dict = None,
        layout: dict = None,
        sub_graph_layout: dict = None,
        sub_graph_data: list = None,
        subplots_kwargs: dict = None,
        **kwargs,
    ):
        """

        :param df: pd.DataFrame

        :param kind_map: dict, subplots graph kind and kwargs
            eg: dict(kind='ScatterGraph', kwargs=dict())

        :param layout: `go.Layout` parameters

        :param sub_graph_layout: Layout of each graphic, similar to 'layout'

        :param sub_graph_data: Instantiation parameters for each sub-graphic
            eg: [(column_name, instance_parameters), ]

            column_name: str or go.Figure

            Instance_parameters:

                - row: int, the row where the graph is located

                - col: int, the col where the graph is located

                - name: str, show name, default column_name in 'df'

                - kind: str, graph kind, default `kind` param, eg: bar, scatter, ...

                - graph_kwargs: dict, graph kwargs, default {}, used in `go.Bar(**graph_kwargs)`

        :param subplots_kwargs: `plotly.tools.make_subplots` original parameters

                - shared_xaxes: bool, default False

                - shared_yaxes: bool, default False

                - vertical_spacing: float, default 0.3 / rows

                - subplot_titles: list, default []
                    If `sub_graph_data` is None, will generate 'subplot_titles' according to `df.columns`,
                    this field will be discarded


                - specs: list, see `make_subplots` docs

                - rows: int, Number of rows in the subplot grid, default 1
                    If `sub_graph_data` is None, will generate 'rows' according to `df`, this field will be discarded

                - cols: int, Number of cols in the subplot grid, default 1
                    If `sub_graph_data` is None, will generate 'cols' according to `df`, this field will be discarded


        :param kwargs:

        """

        self._df = df
        self._layout = layout
        self._sub_graph_layout = sub_graph_layout

        self._kind_map = kind_map
        if self._kind_map is None:
            self._kind_map = dict(kind="ScatterGraph", kwargs=dict())

        self._subplots_kwargs = subplots_kwargs
        if self._subplots_kwargs is None:
            self._init_subplots_kwargs()

        self.__cols = self._subplots_kwargs.get("cols", 2)  # pylint: disable=W0238
        self.__rows = self._subplots_kwargs.get(  # pylint: disable=W0238
            "rows", math.ceil(len(self._df.columns) / self.__cols)
        )

        self._sub_graph_data = sub_graph_data
        if self._sub_graph_data is None:
            self._init_sub_graph_data()

        self._init_figure()

    def _init_sub_graph_data(self):
        """

        :return:
        """
        self._sub_graph_data = []
        self._subplot_titles = []

        for i, column_name in enumerate(self._df.columns):
            row = math.ceil((i + 1) / self.__cols)
            _temp = (i + 1) % self.__cols
            col = _temp if _temp else self.__cols
            res_name = column_name.replace("_", " ")
            _temp_row_data = (
                column_name,
                dict(
                    row=row,
                    col=col,
                    name=res_name,
                    kind=self._kind_map["kind"],
                    graph_kwargs=self._kind_map["kwargs"],
                ),
            )
            self._sub_graph_data.append(_temp_row_data)
            self._subplot_titles.append(res_name)

    def _init_subplots_kwargs(self):
        """

        :return:
        """
        # Default cols, rows
        _cols = 2
        _rows = math.ceil(len(self._df.columns) / 2)
        self._subplots_kwargs = dict()
        self._subplots_kwargs["rows"] = _rows
        self._subplots_kwargs["cols"] = _cols
        self._subplots_kwargs["shared_xaxes"] = False
        self._subplots_kwargs["shared_yaxes"] = False
        self._subplots_kwargs["vertical_spacing"] = 0.3 / _rows
        self._subplots_kwargs["print_grid"] = False
        self._subplots_kwargs["subplot_titles"] = self._df.columns.tolist()

    def _init_figure(self):
        """

        :return:
        """
        self._figure = make_subplots(**self._subplots_kwargs)

        for column_name, column_map in self._sub_graph_data:
            if isinstance(column_name, go.Figure):
                _graph_obj = column_name
            elif isinstance(column_name, str):
                temp_name = column_map.get("name", column_name.replace("_", " "))
                kind = column_map.get("kind", self._kind_map.get("kind", "ScatterGraph"))
                _graph_kwargs = column_map.get("graph_kwargs", self._kind_map.get("kwargs", {}))
                _graph_obj = BaseGraph.get_instance_with_graph_parameters(
                    kind,
                    **dict(
                        df=self._df.loc[:, [column_name]],
                        name_dict={column_name: temp_name},
                        graph_kwargs=_graph_kwargs,
                    ),
                )
            else:
                raise TypeError()

            row = column_map["row"]
            col = column_map["col"]

            _graph_data = getattr(_graph_obj, "data")
            # for _item in _graph_data:
            #     _item.pop('xaxis', None)
            #     _item.pop('yaxis', None)

            for _g_obj in _graph_data:
                self._figure.add_trace(_g_obj, row=row, col=col)

        if self._sub_graph_layout is not None:
            for k, v in self._sub_graph_layout.items():
                self._figure["layout"][k].update(v)

        # NOTE: Use the default theme from plotly version 3.x: template=None
        self._figure["layout"].update(template=None)
        self._figure["layout"].update(self._layout)

    @property
    def figure(self):
        return self._figure



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License. 
Created on 2025/6/25 17:04
Email: yundi.xxii@outlook.com
Description: 注册获取数据的具体实现方法
---------------------------------------------
"""

from typing import Callable

import polars as pl
import quda

FUNC_SQL_LOCAL: Callable[[str, str], pl.LazyFrame] = lambda date, ds_path: quda.sql(
    f"select * from {ds_path} where date='{date}';")


def _sql_local_with_cols(date, ds_path: str, cols: list[str] | None = None):
    if not cols:
        cols = ["*"]
    query = f"select {', '.join(cols)} from {ds_path} where date='{date}';"
    return quda.sql(query)


FUNC_SQL_LOCAL_WITH_COLS: Callable[[str, str, list[str] | None], pl.LazyFrame] = _sql_local_with_cols



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2024/12/9 下午3:07
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""

from pyecharts import options as opts
from matplotlib.colors import to_rgba
from pyecharts.charts import Bar
from pyecharts.globals import ThemeType

class Options:
    # 颜色配置
    colors = [
        'rgba(255, 99, 132, 1)',  # 红色
        'rgba(54, 162, 235, 1)',  # 蓝色
        'rgba(255, 193, 7, 1)',  # 黄色
        'rgba(75, 192, 192, 1)',  # 绿色
        'rgba(153, 102, 255, 1)',  # 紫色
        'rgba(227, 119, 194, 1)',  # 粉色
        'rgba(140, 86, 75, 1)',  # 棕色
        'rgba(255, 187, 120, 1)',  # 橙色
        'rgba(169, 169, 169, 1)',  # 深灰色
        *[f"rgba({int(r*255)}, {int(g*255)}, {int(b*255)}, {int(a)})" for r, g, b, a in [to_rgba(color) for color in Bar(init_opts=opts.InitOpts(theme=ThemeType.MACARONS)).colors]],
    ]

    class Table:
        # 表头设置
        header = dict(font=dict(color="black", size=12),  # dict(size=13, family="Arial, sans-serif", color='#ffffff'),
                      fill_color='gold',  # '#3552a2',
                      align='center',
                      line_color='darkslategray',
                      )
        # 单元格配置
        cell = dict(align='center',
                    line_color='darkslategray',
                    )

    # layout plotly
    layout = dict(template='gridon',
                  hovermode='x unified',
                  hoverlabel=dict(bgcolor='rgba(255, 255, 255, 0.5)'),
                  legend=dict(
                      orientation="h",  # 垂直排列
                      yanchor="top",  # 图例的锚点在上
                      y=1.15,  # 图例的 y 位置
                      xanchor="center",  # 图例的锚点在左
                      x=0.5,  # 图例的 x 位置
                      bordercolor='lightgrey',  # 边框颜色
                      borderwidth=1,  # 边框宽度
                      bgcolor='rgba(255, 255, 255, 0.8)',  # 背景颜色
                  ),
                  # margin=dict(r=0,),
                  xaxis=dict(linecolor="black",
                             linewidth=1,
                             mirror=True,
                             gridcolor='lightgrey',
                             gridwidth=1,
                             griddash='dot',
                             showspikes=True,  # 启用 x 轴的十字光标
                             spikemode='across',  # 十字光标模式
                             spikecolor='black',  # 十字光标颜色
                             spikethickness=1.5,  # 十字光标粗细
                             spikedash='dot',), # 十字光标样式),
                  yaxis=dict(linecolor="black",
                             linewidth=1,
                             mirror=True,
                             gridcolor='lightgrey',
                             gridwidth=1.5,
                             griddash='dot',
                             showspikes=True,  # 启用 x 轴的十字光标
                             spikemode='across',  # 十字光标模式
                             spikecolor='black',  # 十字光标颜色
                             spikethickness=1,  # 十字光标粗细
                             spikedash='dot',),  # 十字光标样式),
                  height=500,
                  )

    @staticmethod
    def get_colors(alpha):
        """更改透明度"""
        new_colors = list()
        for color in Options.colors:
            # 提取 RGB 部分并替换透明度
            rgb_part = color[:-2]  # 去掉最后的透明度部分
            new_color = f"{rgb_part}{alpha})"  # 添加新的透明度
            new_colors.append(new_color)
        return new_colors

    # echarts 悬停框样式
    echarts_opts = dict(
        tooltip_opts=opts.TooltipOpts(
            trigger="axis",
            background_color="rgba(255, 255, 255, 0.5)",  # 背景颜色
            # border_color="black",  # 边框颜色
            border_width=1,  # 边框宽度
            axis_pointer_type="cross",
            textstyle_opts=opts.TextStyleOpts(color="black")  # 文字颜色
        ),
        toolbox_opts=opts.ToolboxOpts(
            is_show=True,
        ),
        datazoom_opts=[
            opts.DataZoomOpts(type_="inside", range_start=0, range_end=100),
            opts.DataZoomOpts(type_="slider", range_start=0, range_end=100)
        ],
    )

    # 悬停十字光标样式
    axispointer_opts = dict(
        axispointer_opts=opts.AxisPointerOpts(
            is_show=True,
            linestyle_opts=opts.LineStyleOpts(
                type_="dashed",
                # color="black",  # 设置十字光标颜色
                width=2         # 设置十字光标宽度
            )
        ),
        # axisline_opts=opts.AxisLineOpts(on_zero_axis_index=1, linestyle_opts=opts.LineStyleOpts(width=1, color="black"))
    )

    @staticmethod
    def get_echarts_options(xaxis_name: str=None, yaxis_name: str=None, title: str=None, hidden: list=None):
        default_opts = dict()
        x_axis = {k: v for k, v in Options.axispointer_opts.items()}
        if xaxis_name is not None:
            x_axis["name"] = xaxis_name
        default_opts["xaxis_opts"] = opts.AxisOpts(**x_axis)
        y_axis = {k: v for k, v in Options.axispointer_opts.items()}
        if yaxis_name is not None:
            y_axis["name"] = xaxis_name
        default_opts["yaxis_opts"] = opts.AxisOpts(**y_axis)
        default_opts.update(Options.echarts_opts)
        if title is not None:
            default_opts["title_opts"] = opts.TitleOpts(title=title)
        if hidden is not None:
            default_opts["legend_opts"] = opts.LegendOpts(selected_map={n: False for n in hidden})
        return default_opts



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/6/8 14:14
@author: ZhangYundi
@email: yundi.xxii@outlook.com
@description: 
---------------------------------------------
"""

from .base import (
    fac_st,
    fac_share,
    fac_cap,
    fac_ipo,
    fac_industry,
    fac_base_quote,
    fac_kline_minute,
    fac_kline_day,
    fac_filter,
    fac_filter_notindex,
)

__all__ = [
    "fac_st",
    "fac_share",
    "fac_cap",
    "fac_ipo",
    "fac_industry",
    "fac_base_quote",
    "fac_kline_minute",
    "fac_kline_day",
    "fac_filter",
    "fac_filter_notindex",
]


# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2024/12/9 下午1:29
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""

import numpy as np
import pandas as pd
import plotly.figure_factory as ff
import plotly.graph_objects as go

from .. import stats
import ylog
from .options import Options


def table(tb_data: pd.DataFrame, highlight_cols: list[str] = None):
    """绘制表格"""
    cols = tb_data.columns
    cellvalues = [[f'<b>{val}</b>' for val in tb_data[cols[0]]]]
    if highlight_cols is not None:
        col_dict = {k: 1 for k in highlight_cols}
        for col in cols[1:]:
            if col in col_dict:
                cellvalues.append([f'<b>{val}</b>' for val in tb_data[col].tolist()])
            else:
                cellvalues.append(tb_data[col])
    else:
        for col in cols[1:]:
            cellvalues.append(tb_data[col].tolist())
    tb = go.Table(
        header=dict(
            values=[f"<b>{col}</b>" for col in tb_data.columns],
            **Options.Table.header,
        ),
        cells=dict(
            values=cellvalues,  # [d.tolist() for g, d in tb_data.items()],
            **Options.Table.cell,
            fill=dict(color=['gold', 'white']),
        ),
    )
    return go.Figure(data=[tb], )


def distplot(data: pd.Series | pd.DataFrame, bin_size):
    """分布图: 每一列都是一组数据集"""
    if isinstance(data, pd.Series):
        data = data.to_frame()
    hist_data = list()
    group_labels = list()
    for col_name, col_data in data.items():
        hist_data.append(col_data.dropna())
        group_labels.append(col_name)
    fig = ff.create_distplot(hist_data, group_labels, bin_size=bin_size, colors=Options.colors)
    fig.update_layout(**Options.layout, )
    if data.index.name == "date":
        fig.update_xaxes(hoverformat='%Y-%m-%d', tickformat="%Y-%m-%d")
    return fig


def violin(data: pd.Series | pd.DataFrame):
    """index是x-axis, columns是datasets"""
    if isinstance(data, pd.Series):
        data = data.to_frame()
    fig = go.Figure()
    x_axis = data.index
    cols = data.columns
    for i, name in enumerate(cols):
        dataset = data[name]
        fig.add_trace(
            go.Violin(
                x=x_axis,
                y=dataset,
                name=name,
                marker_color=Options.colors[i % len(Options.colors)],
                box_visible=True,
                meanline_visible=True,
            )
        )

    fig.update_layout(**Options.layout, )
    if data.index.name == "date":
        fig.update_xaxes(hoverformat='%Y-%m-%d', tickformat="%Y-%m-%d")
    return fig


def bar(data: pd.Series | pd.DataFrame, title: str = None):
    """index是x-axis, columns是datasets"""
    if isinstance(data, pd.Series):
        data = data.to_frame()
    fig = go.Figure()
    x_axis = data.index
    cols = data.columns
    for i, name in enumerate(cols):
        dataset = data[name]
        fig.add_trace(
            go.Bar(
                x=x_axis,
                y=dataset,
                name=name,
                marker_color=Options.colors[i % len(Options.colors)],
            )
        )

    fig.update_layout(**Options.layout, )
    if title is not None:
        fig.update_layout(title=title)
    if data.index.name == "date":
        fig.update_xaxes(hoverformat='%Y-%m-%d', tickformat="%Y-%m-%d")
    return fig


def lines(data: pd.Series | pd.DataFrame):
    if isinstance(data, pd.Series):
        data = data.to_frame()
    fig = go.Figure()
    x_axis = data.index
    cols = data.columns
    for i, name in enumerate(cols):
        dataset = data[name]
        fig.add_trace(
            go.Scatter(
                x=x_axis,
                y=dataset,
                name=name,
                mode="lines",
                marker_color=Options.colors[i % len(Options.colors)],
            )
        )

    fig.update_layout(**Options.layout, )
    if data.index.name == "date":
        fig.update_xaxes(hoverformat='%Y-%m-%d', tickformat="%Y-%m-%d")
    return fig


def nv_plot(data: pd.Series, starting_nv: float = 1.0):
    """
    绘制净值曲线
    Parameters
    ----------
    data: pd.Series
        累计净值, index: date
    starting_nv: float
        初始净值，默认 1

    Returns
    -------

    """
    if not isinstance(data, pd.Series):
        ylog.error(f"Input data must be pandas.Series, not {type(data)}")
        return

    # 收益数据
    ret = data.pct_change().fillna(data[0] / starting_nv - 1)
    # 计算指标
    ann_ret = stats.annual_return(ret)
    cum_ret = data[-1] - starting_nv
    ann_sd = stats.annual_volatility(ret)
    running_max = np.maximum.accumulate(data)
    underwater = (data - running_max) / running_max
    mdd = underwater.min()
    sharpe = ann_ret / ann_sd
    calmar = np.inf if mdd >= 0 else -ann_ret / mdd

    # if isinstance(data, pd.Series):
    #     data = data.to_frame()
    x_axis = data.index
    fig = go.Figure()
    fig.add_trace(
        go.Scatter(
            x=x_axis,
            y=data,
            name=data.name,
            mode="lines",
            # marker_color="rgba(255, 0, 0, 0.8)",
            line=dict(width=2, color='rgba(255, 0, 0, 0.8)', ),
        )
    )
    # 添加最大回撤曲线
    fig.add_trace(
        go.Scatter(x=underwater.index,
                   y=underwater,
                   fill='tozeroy',
                   name='动态回撤',
                   fillcolor='rgba(199, 21, 133, 0.4)',
                   line=dict(width=0.5, color='rgba(199, 21, 133, 0.7)', ),
                   hovertemplate='%{fullData.name}:%{y:.2%}<extra></extra>',
                   yaxis="y2",
                   ),
        # secondary_y=True,
    )
    metric = dict(ann_ret=np.round(ann_ret, 3),
                  cum_ret=np.round(cum_ret, 3),
                  ann_sd=np.round(ann_sd, 3),
                  mdd=np.round(mdd, 5),
                  sharpe=np.round(sharpe, 2),
                  calmar=np.round(calmar, 2),)
    fig.add_annotation(
        xref='paper', yref='paper',
        x=0.02, y=2 / 3,
        xanchor='left', yanchor='middle',
        text=f'年化收益: {ann_ret * 100:.2f}% <br>'
             f'累计收益: {cum_ret * 100:.2f}% <br>'
             f'年化波动: {ann_sd * 100:.2f}% <br>'
             f'最大回撤: {mdd * 100:.2f}% <br>'
             f'sharpe: {sharpe:.2f} <br>'
             f'calmar: {calmar:.2f} <br>',
        font=dict(family='Arial, sans-serif',
                  size=12,
                  color='black'),
        align='left',
        borderwidth=2,
        borderpad=5,
        bordercolor='rgba(220, 220, 220, 0.5)',
        showarrow=False, textangle=0,
    )
    fig.update_layout(**Options.layout, yaxis2=dict(overlaying="y",
                                                    side="right",
                                                    gridcolor='lightgrey',
                                                    gridwidth=1,
                                                    griddash='dot',
                                                    range=[2 * mdd, 0]))
    if data.index.name == "date":
        fig.update_xaxes(hoverformat='%Y-%m-%d', tickformat="%Y-%m-%d")
    return fig, metric



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License.
Created on 2024/11/6 下午7:25
Email: yundi.xxii@outlook.com
---------------------------------------------
"""
import sqlparse
import re

def format_sql(sql_content):
    """将sql语句进行规范化，并去除sql中的注释，输入和输出均为字符串"""
    parse_str = sqlparse.format(sql_content, reindent=True, strip_comments=True)
    return parse_str

def extract_temp_tables(with_clause):
    """从WITH子句中提取临时表名，输出为列表"""
    temp_tables = re.findall(r'\b(\w+)\s*as\s*\(', with_clause, re.IGNORECASE)
    return temp_tables

def extract_table_names_from_sql(sql_query):
    """从sql中提取对应的表名称，输出为列表"""
    table_names = set()
    # 解析SQL语句
    parsed = sqlparse.parse(sql_query)
    # 正则表达式模式，用于匹配表名
    table_name_pattern = r'\bFROM\s+([^\s\(\)\,]+)|\bJOIN\s+([^\s\(\)\,]+)'

    # 用于存储WITH子句中的临时表名
    remove_with_name = []

    # 遍历解析后的语句块
    for statement in parsed:
        # 转换为字符串
        statement_str = str(statement)# .lower()

        # 将字符串中的特殊语法置空
        statement_str = re.sub(r'(substring|extract)\s*\(((.|\s)*?)\)', '', statement_str)

        # 查找匹配的表名
        matches = re.findall(table_name_pattern, statement_str, re.IGNORECASE)

        for match in matches:
            # 提取非空的表名部分
            for name in match:
                if name:
                    # 对于可能包含命名空间的情况，只保留最后一部分作为表名
                    table_name = name.split('.')[-1]
                    # 去除表名中的特殊符号
                    table_name = re.sub(r'("|`|\'|;)', '', table_name)
                    table_names.add(table_name)

        # 处理特殊的WITH语句
        if 'with' in statement_str:
            remove_with_name = extract_temp_tables(statement_str)
    # 移除多余的表名
    if remove_with_name:
        table_names = list(set(table_names) - set(remove_with_name))

    return table_names






# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License.
Created on 2024/7/1 09:44
Email: yundi.xxii@outlook.com
---------------------------------------------
"""
import re
import urllib.parse
from pathlib import Path

from dynaconf import Dynaconf

import ylog
from .parse import extract_table_names_from_sql

USERHOME = Path("~").expanduser() # 用户家目录
NAME = "quda"
DB_PATH = USERHOME / NAME
CONFIG_PATH = DB_PATH / "conf" / "settings.toml"
if not CONFIG_PATH.exists():
    try:
        CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        ylog.error(f"[{NAME}]配置文件生成失败: {e}")
    template_content = f"""[PATHS]
{NAME}="{DB_PATH}" 

# Configuration of database
# [DATABASES.ck]
# urls=["<host1>:<port1>", "<host2>:<port2>",]
# user="xxx"
# password="xxxxxx"
# [DATABASES.mysql]
# url="<host>:<port>"
# user="xxxx"
# password="xxxxxx"

# [UPDATES.mc]
# mod="path.to.your.fetch_module"
# beg_date="2018-01-01"
# update_time="22:00" 
# db_conf="databases.ck"
# mode="auto" # full/auto
# force=false

# 视情况自由增加其他配置\n"""
    with open(CONFIG_PATH, "w") as f:
        f.write(template_content)
    ylog.info(f"[{NAME}]生成配置文件: {CONFIG_PATH}")

def get_settings():
    try:
        return Dynaconf(settings_files=[CONFIG_PATH])
    except Exception as e:
        ylog.error(f"[{NAME}]读取配置文件失败: {e}")
        return {}

# 读取配置文件覆盖
_settiings = get_settings()
if _settiings is not None:
    DB_PATH = _settiings.get(f"paths.{NAME}", "")
    DB_PATH = Path(DB_PATH)


# ======================== 本地数据库 catdb ========================
def tb_path(tb_name: str) -> Path:
    """
    返回指定表名 完整的本地路径
    Parameters
    ----------
    tb_name: str
       表名，路径写法: a/b/c
    Returns
    -------
    pathlib.Path
        full_abs_path: pathlib.Path
        完整的本地绝对路径 $DB_PATH/a/b/c
    """
    return Path(DB_PATH, tb_name)


def put(df, tb_name: str, partitions: list[str] | None = None, abs_path: bool = False):
    """
    将一个DataFrame写入到指定名称的表格目录中，支持分区存储。

    该函数负责将给定的DataFrame（df）根据提供的表名（tb_name）写入到本地文件系统中。
    如果指定了分区（partitions），则会按照这些分区列将数据分割存储。此外，可以通过abs_path参数
    指定tb_name是否为绝对路径。如果目录不存在，会自动创建目录。

    Parameters
    ----------
    df: polars.DataFrame
    tb_name: str
        表的名称，用于确定存储数据的目录
    partitions: list[str] | None
        指定用于分区的列名列表。如果未提供，则不进行分区。
    abs_path: bool
        tb_name是否应被视为绝对路径。默认为False。

    Returns
    -------

    """
    if df is None:
        ylog.warning(f"[{NAME}] put failed: input data is None.")
        return
    if df.is_empty():
        ylog.warning(f"[{NAME}] put failed: input data is empty.")
        return
    if not abs_path:
        tbpath = tb_path(tb_name)
    else:
        tbpath = Path(tb_name)
    if not tbpath.exists():
        tbpath.mkdir(parents=True, exist_ok=True)
    if partitions is not None:
        df.write_parquet(tbpath, partition_by=partitions)
    else:
        df.write_parquet(tbpath / "data.parquet")

def has(tb_name: str) -> bool:
    """
    判定给定的表名是否存在
    Parameters
    ----------
    tb_name: str

    Returns
    -------

    """
    return tb_path(tb_name).exists()

def sql(query: str, abs_path: bool = False, lazy: bool = True):
    """
    sql 查询，从本地paquet文件中查询数据

    Parameters
    ----------
    query: str
        sql查询语句
    abs_path: bool
        是否使用绝对路径作为表路径。默认为False
    lazy: bool
        惰性模式
    Returns
    -------

    """
    import polars as pl

    tbs = extract_table_names_from_sql(query)
    convertor = dict()
    for tb in tbs:
        if not abs_path:
            db_path = tb_path(tb)
        else:
            db_path = tb
        format_tb = f"read_parquet('{db_path}/**/*.parquet')"
        convertor[tb] = format_tb
    pattern = re.compile("|".join(re.escape(k) for k in convertor.keys()))
    new_query = pattern.sub(lambda m: convertor[m.group(0)], query)
    if not lazy:
        return pl.sql(new_query).collect()
    return pl.sql(new_query)


def read_mysql(query: str, db_conf: str = "DATABASES.mysql"):
    """
    从MySQL数据库中读取数据。
    Parameters
    ----------
    query: str
        查询语句
    db_conf: str
        对应的配置 $DB_PATH/conf/settings.toml
    Returns
    -------
    polars.DataFrame
    """
    import polars as pl
    try:
        db_setting = get_settings().get(db_conf, {})
        required_keys = ['user', 'password', 'url']
        missing_keys = [key for key in required_keys if key not in db_setting]
        if missing_keys:
            raise KeyError(f"Missing required keys in database config: {missing_keys}")

        user = urllib.parse.quote_plus(db_setting['user'])
        password = urllib.parse.quote_plus(db_setting['password'])
        uri = f"mysql://{user}:{password}@{db_setting['url']}"
        return pl.read_database_uri(query, uri)

    except KeyError as e:
        raise RuntimeError("Database configuration error: missing required fields.") from e
    except Exception as e:
        raise RuntimeError(f"Failed to execute MySQL query: {e}") from e


def read_ck(query: str, db_conf: str = "DATABASES.ck"):
    """
    从Clickhouse集群读取数据。
    Parameters
    ----------
    query: str
        查询语句
    db_conf: str
        对应的配置 $DB_PATH/conf/settings.toml
    Returns
    -------
    polars.DataFrame
    """
    import clickhouse_df
    try:
        db_setting = get_settings().get(db_conf, {})
        required_keys = ['user', 'password', 'urls']
        missing_keys = [key for key in required_keys if key not in db_setting]
        if missing_keys:
            raise KeyError(f"Missing required keys in database config: {missing_keys}")

        user = urllib.parse.quote_plus(db_setting['user'])
        password = urllib.parse.quote_plus(db_setting['password'])

        with clickhouse_df.connect(db_setting['urls'], user=user, password=password):
            return clickhouse_df.to_polars(query)

    except KeyError as e:
        raise RuntimeError("Database configuration error: missing required fields.") from e
    except Exception as e:
        raise RuntimeError(f"Failed to execute ClickHouse query: {e}") from e



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/5/30 16:34
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""

_BASE_LOCAL = "jydata"
TB_CALENDAR = f"{_BASE_LOCAL}/交易日历"
TB_SECU_MAIN = f"{_BASE_LOCAL}/证券主表"
TB_SHARES_INFO = f"{_BASE_LOCAL}/股本信息"
TB_LIST_STATUS = f"{_BASE_LOCAL}/上市信息"
TB_SPECIAL_TRADE = f"{_BASE_LOCAL}/特别处理"
TB_INDUSTRY_SW = f"{_BASE_LOCAL}/申万行业"
TB_ADJ_FACTOR = f"{_BASE_LOCAL}/复权因子"


_MC_LOCAL = "mc"
TB_STOCK_KLINE_DAY = f"{_MC_LOCAL}/stock_kline_day"
TB_STOCK_KLINE_MINUTE = f"{_MC_LOCAL}/stock_kline_minute"
TB_STOCK_TICK = f"{_MC_LOCAL}/stock_tick"
TB_INDEX_KLINE_DAY = f"{_MC_LOCAL}/index_kline_day"
TB_INDEX_KLINE_MINUTE = f"{_MC_LOCAL}/index_kline_minute"



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License. 
Created on 2025/7/9 14:02
Email: yundi.xxii@outlook.com
Description: 
---------------------------------------------
"""

import polars as pl

class OrderConfig:

    def __init__(self, signal: str, order_unit: int = 100, amount_unit: float = -1):
        self.signal = signal
        self.order_unit = order_unit
        self.amount_unit = amount_unit

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(signal={self.signal}, order_unit={self.order_unit}, amount_unit={self.amount_unit})"

    def __repr__(self):
        return self.__str__()

def create_open_order(signal_data: pl.LazyFrame, open_config: OrderConfig) -> pl.LazyFrame:
    """
    创建开仓单
    Parameters
    ----------
    signal_data: polars.LazyFrame
        信号数据, 必须包含 `price` 列 以及 order_config.signal 列
    open_config: OrderConfig
        订单配置
    Returns
    -------
    polars.LazyFrame
    """
    assert open_config.signal in signal_data.collect_schema().names(), f"create_open_order.signal_data 缺少 `{open_config.signal}` 列"
    scalar = pl.lit(1) if open_config.amount_unit < 0 else (open_config.amount_unit / pl.col("price") / open_config.order_unit).ceil()
    return (
        signal_data
        .with_columns(
            (pl.col(open_config.signal) * scalar * open_config.order_unit).fill_null(0).cast(pl.Int32).alias(f"order_qty_{open_config.signal}")
        )
    )

def create_open_orders(signal_data: pl.LazyFrame, open_configs: list[OrderConfig]) -> pl.LazyFrame:
    """创建多个信号的开仓单"""
    open_orders = [create_open_order(signal_data, open_config) for open_config in open_configs]
    open_orders = pl.concat(open_orders, how="diagonal")
    open_orders = (
        open_orders
        .drop("^.*open_signal.*$", "^.*cover_signal.*$", strict=False)
        .with_columns(selected_signal=pl.sum_horizontal("^.*signal.*$"),
                      order_qty=pl.sum_horizontal("^.*order_qty.*$"),)
        .with_columns(open_signal=pl.col("order_qty").sign(),
                      direction=pl.col("order_qty").sign())
        .with_columns(order_qty=pl.col("order_qty").abs(),
                      is_open=pl.col("open_signal").abs())
    )
    date_window_spec = dict(partition_by=["date", "asset"], order_by="time")
    open_orders = (
        open_orders
        .filter(pl.col("is_open") > 0)
        .with_columns(last_direction=pl.col("direction").shift(1).over(**date_window_spec).fill_null(0))
        .with_columns(
            action_id=pl.when(pl.col("direction") != pl.col("last_direction")).then(1).otherwise(0).cum_sum().over(
                **date_window_spec))
        .drop("last_direction")
    )

    return open_orders

def cover_order_rs(open_order: pl.LazyFrame) -> pl.LazyFrame:
    """平仓单: 信号反转 reversed signal"""
    cols = open_order.columns
    check_fields = ["direction", "order_qty"]
    for field in check_fields:
        assert field in cols, f"cover_order_rs.open_order 缺少 `{field}` 列"
    date_window_spec = dict(partition_by=["date", "asset"], order_by="time")
    open_window_spec = dict(partition_by=["date", "asset", "action_id"], order_by="time")
    open_order = (
        open_order
        .with_columns(cum_open_qty=pl.col("order_qty").cum_sum().over(**open_window_spec).fill_null(0))
        .with_columns(last_cum_open_qty=pl.col("cum_open_qty").shift(1).over(**date_window_spec).fill_null(0),
                      last_direction=pl.col("direction").shift(1).over(**date_window_spec).fill_null(0))
    )
    cover_order = (
        open_order
        .with_columns(order_qty=pl.when(pl.col("last_direction") != 0,
                                        pl.col("direction") != pl.col("last_direction"))
                      .then(pl.col("last_cum_open_qty"))
                      .otherwise(None))
        .drop_nulls(subset=["order_qty"])
        .with_columns(is_open=pl.lit(0))
        .select(cols)
    )
    return cover_order

def cover_order_t0(order_data: pl.LazyFrame) -> pl.LazyFrame:
    """
    平仓单: t0收盘平掉所有敞口
    """
    cols = order_data.columns
    check_fields = ["price", "direction", "order_qty"]
    for field in check_fields:
        assert field in cols, f"`cover_order_t0.order_data` 缺少必要字段 {field}"
    # 当天的累计敞口
    total_qty = order_data.group_by("date", "asset").agg(total_qty=(pl.col("order_qty") * pl.col("direction")).sum())
    cover_order = (
        total_qty
        .with_columns(direction=-pl.col("total_qty").sign(),
                      order_qty=pl.col("total_qty").abs(), )
        .filter(pl.col("order_qty") > 0)
        .drop("total_qty")
    )
    return cover_order

def create_order(signal_data: pl.DataFrame | pl.LazyFrame,
                 open_configs: list[OrderConfig],
                 cover_configs: list[OrderConfig],
                 cover=True) -> pl.LazyFrame:
    """
    根据信号进行开平仓

    Parameters
    ----------
    signal_data: polars.DataFrame | polars.LazyFrame
    open_configs: list[OrderConfig]
    cover_configs: list[OrderConfig]
    cover: bool

    Returns
    -------
    polars.DataFrame | polars.LazyFrame
    """

    signal_data = signal_data.lazy()
    assert "price" in signal_data.collect_schema().names(), f"create_order.signal_data 缺少 `price` 列"
    open_orders = create_open_orders(signal_data, open_configs)
    open_orders_shadow = create_open_orders(signal_data, cover_configs)
    cover_orders = cover_order_rs(open_orders_shadow)
    cover_orders = (
        cover_orders
        .with_columns(is_close=pl.lit(1),
                      cover_signal=pl.col("selected_signal"))
    )
    orders = pl.concat([open_orders, cover_orders], how="diagonal").sort("date", "time", "asset", "is_open")
    date_window_spec = dict(partition_by=["date", "asset"], order_by="time")
    action_window_spec = dict(partition_by=["date", "asset", "action_id"], order_by="time")
    orders = (
        orders
        .drop("selected_signal")
        .with_columns(last_open_signal=pl.col("open_signal").shift(1).over(**date_window_spec))
        .filter(~pl.all_horizontal(pl.col("open_signal").is_null(),
                                   pl.col(
                                       "last_open_signal").is_null(), ))  # open_signal 为 Null，则必是平仓单，如果last_open_direction 为Null, 说明这是无效平仓单，剔除
        .with_columns(pl.col("open_signal").fill_null(0),
                      pl.col("cover_signal").fill_null(0),
                      pl.col("last_open_signal").fill_null(0))
        .filter(pl.col("last_open_signal") != pl.col("cover_signal"), pl.col("open_signal") != pl.col("cover_signal"))
        .with_columns(last_cover_signal=pl.col("cover_signal").shift(1).over(**date_window_spec))
        .with_columns(action_id=pl.when(pl.col("last_cover_signal").abs() > 0,
                                        pl.col("open_signal") != pl.col("last_open_signal"))
                      .then(1)
                      .otherwise(0)
                      .cum_sum()
                      .over(**date_window_spec))
    )
    # 修复平仓单量
    orders = (
        orders
        .with_columns(open_qty=pl.when(pl.col("is_close") == 1).then(None).otherwise(pl.col("order_qty")))
        .with_columns(cover_qty=pl.col("open_qty").cum_sum().forward_fill().over(**action_window_spec))
        .with_columns(order_qty=pl.when(pl.col("is_open") == 1).then(pl.col("open_qty")).otherwise(pl.col("cover_qty")))
        .drop("open_qty", "cover_qty")
    )
    if cover:
        latest_signal_data = signal_data.group_by("date", "asset").agg(pl.all().last())
        cover_orders = cover_order_t0(orders)
        cover_orders = (
            cover_orders
            .select("date", "asset", "direction", "order_qty")
            .join(latest_signal_data.select("date", "time", "asset", "price"),
                  on=["date", "asset"],
                  how="left")
            .with_columns(
                is_open=pl.lit(0),
                is_close=pl.lit(1),
            )
        )
        orders = pl.concat([orders, cover_orders], how="diagonal").sort("date", "time", "asset", "is_open")
        orders = orders.with_columns(action_id=pl.col("action_id").forward_fill().over(**date_window_spec))
    return orders

def summary_order_uc(order_data: pl.LazyFrame, limit: bool = True, fee: float = 5e-4, tax: float = 5e-4) -> pl.LazyFrame:
    """订单数据统计: 添加`pnl`列以及过滤只留下开仓信号"""
    action_window_spec = dict(partition_by=["date", "asset", "action_id"], order_by="time")
    window_spec = dict(partition_by=["date", "asset"], order_by="time")

    order_data = (
        order_data
        .with_columns(
            cover_price=pl.when(pl.col("is_open") == 0).then(pl.col("price")).otherwise(None).backward_fill().over(
                **action_window_spec),
            cover_time=pl.when(pl.col("is_open") == 0).then(pl.col("time")).otherwise(None).backward_fill().over(
                **action_window_spec))
    )
    if limit:
        assert "limit_qty" in order_data.columns, f"创建订单数据源必须包含 `limit_qty` 列"
        # 根据 limit_qty 修改订单数量
        order_data = (order_data
                      .filter(pl.col("is_open") == 1)
                      .sort("date", "time", "asset")
                      .with_columns(order_qty=pl.col("order_qty").clip(upper_bound=pl.col("limit_qty")))
                      .with_columns(cum_qty=pl.col("order_qty").cum_sum().over("date", "asset", order_by="time"))
                      .with_columns(cum_qty=pl.col("cum_qty").clip(upper_bound=pl.col("limit_qty")))
                      .with_columns(order_qty=pl.col("cum_qty").diff().over(**window_spec).fill_null(pl.col("order_qty")))
                      .filter(pl.col("order_qty") > 0)
                      )

    order_data = (
        order_data
        .with_columns(pnl=pl.when(pl.col("direction") > 0)
                      .then(pl.col("cover_price") * (1 - fee - tax) - pl.col("price") * (1 + fee))
                      .otherwise(pl.col("price") * (1 - fee - tax) - pl.col("cover_price") * (1 + fee)))
        .with_columns(pnl=pl.col("pnl") * pl.col("order_qty"))
        .filter(pl.col("is_open") == 1)
    )
    return order_data






# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License.
Created on 2025/6/25 15:05
Email: yundi.xxii@outlook.com
Description:
---------------------------------------------
"""

from collections.abc import Sequence
from typing import Literal
import numpy as np
import ylog
import polars as pl

def get_timedelta_names(columns: Sequence[str], ) -> list[str]:
    """
    Return a list of column names that can be successfully parsed into <pandas>.Timedelta object.

    Parameters
    ----------
    columns: Sequence[str]
        A Sequence of strings representing column names to be evaluated for Timedelta

    Returns
    -------
    list[str]

    Examples
    --------
    >>> get_timedelta_names(['1s', '2min', 'invalid', '3H'])
    ['1s', '2min', '3H']

    """
    from pandas import Timedelta
    def _is_valid_timedelta(s: str) -> bool:
        try:
            Timedelta(s)
            return True
        except ValueError:
            return False

    return [col for col in columns if _is_valid_timedelta(col)]


def extract_feature_names(columns: Sequence[str], ) -> list[str]:
    """
    Extract a list of valid feature column names by excluding index-like(date/time/asset)、price-like(price)、time-like columns.
    Parameters
    ----------
    columns: Sequence[str]
        A sequence of strings representing all available column names.

    Returns
    -------
    list[str]

    Examples
    --------
    >>> extract_feature_names(['date', 'time', 'asset', 'open', 'high', '1s', 'volume', 'price'])
    ['open', 'high', 'volume']

    """
    return_cols = get_timedelta_names(columns)
    exclude = {"date", "time", "asset", 'price', *return_cols}
    return [col for col in columns if col not in exclude]

def preprocess_features(df: pl.DataFrame,
                        target_col: str,
                        missing_threshold: float = 0.95,
                        standardize: bool = True,
                        var_threshold: float = 0.01,
                        index: tuple[str] = ("date", "time", "asset")) -> pl.DataFrame:
    """
    预处理阶段: 去除缺失值过高和低方差/常量特征
    Parameters
    ----------
    df: pl.DataFrame
        输入数据
    target_col: str
        目标列名称
    var_threshold: float
        方差阈值，低于此值的特征将被删除
    missing_threshold: float
        缺失值比例阈值，高于此值的特征将被删除
    standardize: bool
        是否对特征进行标准化(z-score)
    index: tuple[str]
        索引列, 预处理步骤会跳过这些列

    Returns
    -------
    df_cleaned: pl.DataFrame
        清洗后的DataFrame
    """
    ylog.info("Preprocessing...")
    df = df.fill_nan(None)
    skip_cols = set(index)
    if target_col:
        skip_cols.add(target_col)
    features = [col for col in df.columns if col not in skip_cols]

    def filter_cols(data):
        """
        删除常量列
        """
        return (
            data[0]
            .transpose(include_header=True,
                       column_names=["cond"],
                       header_name="feat_name")
            .filter(pl.col("cond"))
            .drop_nulls("cond")
            ["feat_name"].to_list()
        )

    # 删除常量列
    feat_consts = df[features].select(pl.col(c).drop_nulls().n_unique() for c in features)
    not_const_exprs = filter_cols(feat_consts > 1)

    # 处理 inf
    df = (
        df
        .with_columns(pl.col(c)
                      .cast(pl.Float32)
                      .replace([np.inf, -np.inf], None) for c in not_const_exprs
                      )
    )
    # 删除缺失值比例过高的列
    feat_null_ratio = df[not_const_exprs].null_count() / df.height
    valid_exprs = filter_cols(feat_null_ratio <= missing_threshold)
    # 标准化(可选)
    if standardize:
        df = (
            df
            .select(
                *skip_cols,
                *[(pl.col(c) - pl.col(c).mean()) / pl.col(c).std() for c in valid_exprs],
            )
        )
    # 删除常量或低方差特征
    # 计算各列方差
    if standardize:
        feat_var = df[valid_exprs].select(pl.all().var())
        high_var_exprs = filter_cols(feat_var > var_threshold)
        return df.select(*index, *high_var_exprs, target_col)
    return df.select(*index, *valid_exprs, target_col)

def fast_rfe(df,
             target_col: str,
             tail_label: tuple[float] = (5, 95),
             index: tuple[str] = ("date", "time", "asset"),
             missing_threshold: float = 0.4,
             standardize: bool = True,
             var_threshold: float = 0.01,
             fill_null_strategy: Literal[
                                     "forward", "backward", "mean", "zero", "max", "min", "one"] | None = "forward"):
    """特征筛选"""
    from lightgbm import LGBMRegressor
    if hasattr(df, "__call__"):
        df = df()
    if isinstance(df, pl.LazyFrame):
        df = df.collect()
    # 预处理
    df = preprocess_features(df,
                             target_col=target_col,
                             index=index,
                             missing_threshold=missing_threshold,
                             standardize=standardize,
                             var_threshold=var_threshold,
                             )
    # 处理空值
    skip_cols = set(index) if index else set()
    skip_cols.add(target_col)
    feat_names = [feat for feat in df.columns if feat not in skip_cols]
    fill_null_over_spec = {"partition_by": "asset", "order_by": ["date", "time"]}
    df = (
        df
        .with_columns(pl.col(c)
                      .fill_null(strategy=fill_null_strategy)
                      .over(**fill_null_over_spec)
                      for c in feat_names)
        .with_columns(pl.col(target_col)
                      .fill_null(strategy=fill_null_strategy)
                      .over(**fill_null_over_spec))
        .drop_nulls()
    )

    X = df[feat_names].to_numpy()
    y = df[target_col].to_numpy().ravel()

    # 根据label来减少样本数量
    q_low, q_high = np.percentile(y, tail_label)
    mask = (y < q_low) | (y > q_high)
    X = X[mask]
    y = y[mask]

    ylog.info("Running Fast-RFE...")
    model = LGBMRegressor(n_estimators=100, random_state=42, )
    model.fit(X, y, )
    importance = model.feature_importances_
    # 排序并且提出最不重要的前k个特征
    sorted_feats = sorted(zip(feat_names, importance), key=lambda x: x[1], reverse=True)
    selected_feats = [f for f, s in sorted_feats if s > 0]
    # todo: 相关性 > 0.9 的特征保留重要性大的
    # corr_df = df[selected_feats].corr()
    # for feat in selected_feats:
    return selected_feats


def is_notebook():
    import sys
    return "ipykernel" in sys.modules

def hydra_wrapper(config_path, config_name):
    """将 hydra.main 装饰器转为方法调用"""
    import os
    from hydra import main
    abs_config_path = os.path.abspath(config_path)
    if not os.path.exists(abs_config_path):
        ylog.error(f"Miss {abs_config_path}")
        return
    return main(version_base=None, config_path=abs_config_path, config_name=config_name)


# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/6/6 17:15
@author: ZhangYundi
@email: yundi.xxii@outlook.com
@description: 因子数据补齐模块
---------------------------------------------
"""
from collections import defaultdict
from glob import glob
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import unquote

import pandas as pd
import polars as pl
import xcals
import ygo
import ylog

from .consts import *
from .errors import FactorGetError
from ..core.store import database

if TYPE_CHECKING:
    from .core import Factor


class MissingChecker:
    @staticmethod
    def check_date(fac: 'Factor', beg_date, end_date) -> list[str]:
        """检验数据缺失的日期"""
        dateList = xcals.get_tradingdays(beg_date=beg_date, end_date=end_date)
        if not database.has(fac.tb_name):
            return dateList
        # 查询本地有数据的日期
        fac_path = database.tb_path(fac.tb_name)
        schema = pl.scan_parquet(fac_path).collect_schema()
        columns = schema.names()
        cond = " OR ".join([f"'{col}' IS NOT NULL" for col in columns])
        sql = f"""SELECT date
                    FROM {fac.tb_name}
                WHERE date BETWEEN '{beg_date}' AND '{end_date}'
                    AND ({cond})
                GROUP BY date 
                HAVING count() > 0;"""
        exist_dateList = database.sql(sql, lazy=False)["date"].cast(pl.Utf8).to_list()
        return set(dateList) - set(exist_dateList)

    @staticmethod
    def check_times(fac: 'Factor', date: str, beg_time: str, end_time: str, freq: str) -> set:
        """
        获取指定日期缺失的时间段
        Parameters
        ----------

        Returns
        -------

        """
        tb_name = fac.tb_name
        tb_path = database.tb_path(tb_name) / f"date={date}"
        partition_dirs = glob(str(tb_path / "time=*" / "*.parquet"))
        exists_times = set()
        for p in partition_dirs:
            rel_path = Path(p).relative_to(tb_path)
            parts = dict(part.split("=", 1) for part in rel_path.parts if "=" in part)
            exists_times.add(unquote(parts["time"]))
        if beg_time == end_time:
            need_times = [beg_time, ]
        else:
            need_times = xcals.get_tradingtime(beg_time, end_time, freq)
        return set(need_times) - set(exists_times)

    @staticmethod
    def check_datetimes(fac: 'Factor',
                        beg_date: str,
                        end_date: str,
                        beg_time: str,
                        end_time: str,
                        freq: str) -> defaultdict[set]:
        """
        获取本地已存在的日期时间
        Parameters
        ----------

        Returns
        -------

        """
        missing_datetimes = defaultdict(set)
        for date in xcals.get_tradingdays(beg_date, end_date):
            missing_times = MissingChecker.check_times(fac, date=date, beg_time=beg_time,
                                                       end_time=end_time, freq=freq)
            if missing_times:
                missing_datetimes[date] = missing_times
        return missing_datetimes


class DataCompleter:
    """数据补齐器"""

    @staticmethod
    def _get_value_firsttime(fac: 'Factor', date: str, ) -> pl.DataFrame:
        """
        第一次落数据: 当本地没有数据或者数据都为空


        Parameters
        ----------
        fac : Factor
            因子对象，包含因子计算函数和其他相关信息。
        date : str
            日期，用于指定因子计算的日期，格式为 yyyy-mm-dd

        Returns
        -------
        data : pl.DataFrame | None
            处理后的因子计算结果数据

        Raises
        ------
        Exception
            如果因子计算函数返回的数据为空或数据类型不符合要求，则抛出异常。
        Exception
            如果因子计算结果中缺少必要的 `asset` 列，则抛出异常。
        """

        data = ygo.delay(fac.fn)(this=fac, date=date)()
        e = FactorGetError.new_error(fac, date, fac.end_time, "Empty Value.")
        if data is None:
            return fac.tb_name, None, e
        if not (isinstance(data, (pl.DataFrame, pd.Series, pd.DataFrame))):
            raise Exception("因子计算函数需要返回 polars.DataFrame | pandas.Series | pandas.DataFrame")
        if isinstance(data, (pd.DataFrame, pd.Series)):
            index_levs = data.index.nlevels
            if index_levs == 1:
                assert FIELD_ASSET == data.index.name, f"因子计算结果index中必须包含`{FIELD_ASSET}`"
            else:
                assert FIELD_ASSET in data.index.names, f"因子计算结果index中必须包含`{FIELD_ASSET}`"
            data = pl.from_pandas(data.reset_index())
        if FIELD_ASSET not in data.columns:
            if data.is_empty():
                return fac.tb_name, None, e
            raise Exception(f"因子计算函数返回值中必须包含`{FIELD_ASSET}`列")
        val_fields = data.drop(INDEX, strict=False).columns

        data = data.unique().fill_nan(None)
        if data.drop_nulls().is_empty():
            return fac.tb_name, None, e
        if FIELD_DATE not in data.columns:
            data = data.with_columns(pl.lit(date).alias(FIELD_DATE))
        data = data.with_columns(pl.lit(fac.end_time).alias(FIELD_TIME))
        data = data.select(*INDEX, *val_fields, )
        return fac.tb_name, data.sort(INDEX), None

    @staticmethod
    def complete(missing_config: list[tuple['Factor', dict[str, set]]], n_jobs: int):
        if missing_config:
            with ygo.pool(n_jobs=n_jobs) as go:
                for (fac, missing_datetimes) in missing_config:
                    for date, timeset in missing_datetimes.items():
                        for time_ in timeset:
                            job_name = f"[{fac.name} completing]"
                            if fac.type == TYPE_REALTIME:
                                fac = fac(end_time=time_)
                            go.submit(DataCompleter._get_value_firsttime, job_name=job_name)(fac=fac,
                                                                                             date=date, )

                dataset = defaultdict(list)
                for (tb_name, data, e) in go.do():
                    if data is None:
                        ylog.warning(e)
                        continue
                    dataset[tb_name].append(data)
                if dataset:
                    for tb_name, data_list in dataset.items():
                        data = pl.concat(data_list)
                        database.put(data, tb_name=tb_name, partitions=[FIELD_DATE, FIELD_TIME])



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License. 
Created on 2025/6/25 15:05
Email: yundi.xxii@outlook.com
Description: 数据加载器，支持分批加载和流水线处理
---------------------------------------------
"""
from collections.abc import Sequence

from omegaconf import DictConfig
from sklearn.base import BaseEstimator
from sklearn.pipeline import Pipeline
from tqdm.auto import tqdm

from ..utils import is_notebook, hydra_wrapper


class DataLoader(BaseEstimator):
    """
    数据加载器

    该类封装了 `sklearn.pipeline.Pipeline`，支持以批处理方式执行多个数据转换步骤，
    并能将结果合并为一个 `polars.DataFrame`。适用于大规模数据集的流式处理。

    Attributes
    ----------
    pipe : sklearn.pipeline.Pipeline
        内部使用的数据处理流水线对象。
    """

    def __init__(self,
                 steps: list):

        """
        初始化 DataLoader 实例。

        Parameters
        ----------
        steps : list of (str, transformer) tuples
            流水线中的各个处理步骤，格式与 sklearn 的 Pipeline 相同。

            示例：

            >>> steps = [('step1', Transformer1()), ('step2', Transformer2())]

        See Also
        --------
        sklearn.pipeline.Pipeline : sklearn 中用于构建数据处理流程的标准类。

        """
        pipe_steps = list()
        if steps:
            for name, step in steps:
                if isinstance(step, type):
                    pipe_steps.append((name, step()))
                else:
                    pipe_steps.append((name, step))
        self.pipe = Pipeline(pipe_steps)
        self._html_repr = self.pipe._html_repr

    def fetch(self,
              iterables: Sequence,
              batch_size: int = -1,
              eager: bool = True):
        """
        分批次加载并处理数据。

        将输入的可迭代对象按 `batch_size` 分批加载。每个单位(iterables.item)的数据会通过 `Pipeline` 进行转换，
        最终合并成一个 `polars.DataFrame` 返回。

        Parameters
        ----------
        iterables: Sequence
            可迭代对象
        batch_size: int
            每次处理的数据量大小。默认为 -1，表示一次性返回所有数据。
        eager: bool

        Yields
        -------
        polars.DataFrame | polars.LazyFrame
            处理后的数据批次，合并为一个 DataFrame 后逐个返回。

        Examples
        --------
        >>> from sklearn.preprocessing import FunctionTransformer
        >>> import pandas as pd
        >>>
        >>> steps = [('transformer', FunctionTransformer(lambda x: pd.DataFrame({"x": [x]})))]
        >>> loader = DataLoader(steps)
        >>> data = [1, 2, 3, 4, 5]
        >>> for batch in loader.fetch(data, batch_size=2):
        ...     print(batch)

        """

        from polars import concat, LazyFrame

        batch_size = batch_size if batch_size > 0 else len(iterables)
        batch_list = [iterables[i: i + batch_size] for i in range(0, len(iterables), batch_size)]
        with tqdm(total=len(batch_list), desc=f"{self.__class__.__name__}.fetching", leave=False) as pbar:
            for batch in batch_list:
                lfList = [self.pipe.fit_transform(batch[i]) for i in
                          tqdm(range(len(batch)), desc="Tranforming batch lazyframe", leave=False)]
                lf = concat(lfList)
                if isinstance(lf, LazyFrame) and eager:
                    lf = lf.collect()
                pbar.update()
                yield lf

    def set_params(self, **params):
        """允许通过类似 sklearn 的 params__value设置参数"""
        self.pipe.set_params(**params)

    def get_params(self, deep=True):
        """返回所有步骤的参数，用于 mlflow logging 或者 gridsearch cv"""
        return self.pipe.get_params(deep=deep)

    @classmethod
    def initialize(cls, cfg: DictConfig) -> 'DataLoader':
        """
        从 解析出来的 初始化 DataLoader
        """

        from hydra.utils import instantiate

        # 构建 data pipeline
        data_steps = []
        data_steps = instantiate(cfg.data)
        data_steps = [(name, step) for d in data_steps for name, step in d.items()]
        return cls(data_steps)

    @classmethod
    def from_conf(cls, config_path: str, config_name: str) -> 'DataLoader':
        if is_notebook():
            print(f"""Run in notebook, please use hydra to run.
::code::
from hydra import initialize, compose
with initialize(version_base=None, config_path='{config_path}'):
    cfg = compose(config_name='{config_name}')
dl = quda.ml.DataLoader.initialize(cfg)
""")
            return

        hydra_main = hydra_wrapper(config_path, config_name)
        if hydra_main is not None:
            return hydra_main(cls.initialize)()



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/6/8 14:29
@author: ZhangYundi
@email: yundi.xxii@outlook.com
@description: 
---------------------------------------------
"""
import time

import polars as pl
from quda.data.zoo import base
import ylog
import ygo

from quda.data.zoo.base import fac_components


# from quda.factor.core import FactorContext

def test_components():
    test_date = "2025-05-06"
    w = fac_components(index_code="000852").get_value(test_date)
    ylog.info(w)

def test_base_quote():
    test_date1 = "2025-05-06"
    test_date2 = "2025-07-04"
    data1 = base.fac_base_quote.get_value(test_date1)
    data2 = base.fac_base_quote(env="rt").get_value(test_date2)
    ylog.info(data1.head())
    ylog.info(data2.head())

def test_filter():
    test_date = "2025-05-06"
    data = base.fac_filter.get_value(test_date)
    ylog.info(data.filter(pl.col("cond") > 0))

def test_filter_notindex():
    test_date = "2025-05-06"
    data = base.fac_filter_notindex(index_codes=["000016", ]).get_value(test_date)
    ylog.info(data.filter(pl.col("cond").is_null()))

def test_get_value():
    test_date = "2025-05-06"
    data = base.fac_kline_minute.get_value(date=test_date, time="09:31:00")
    # data = base.fac_filter.get_value(date=test_date, time="09:00:00")
    ylog.info(data)



if __name__ == '__main__':
    # print(FactorContext.__dataclass_fields__.keys())
    # print(ygo.fn_info(ygo.delay(base.fac_prev_close.fn)(env="rt")))
    # test_base_quote()
    # test_filter()
    # test_components()
    # test_filter_notindex()
    test_get_value()


# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License. 
Created on 2025/7/9 15:03
Email: yundi.xxii@outlook.com
Description: 
---------------------------------------------
"""

import polars as pl
from .order import OrderConfig, create_order, summary_order_uc
import pandas as pd

def from_signal(signal_data: pl.LazyFrame,
                open_configs: list[OrderConfig],
                cover_configs: list[OrderConfig],
                cover: bool = True,
                limit: bool = True,
                fee: float = 5e-4,
                tax: float = 5e-4,
                show_notebook: bool = True):

    order_oss = create_order(signal_data, open_configs=open_configs, cover_configs=cover_configs, cover=cover)
    order_uc = summary_order_uc(order_oss, limit=limit, fee=fee, tax=tax, )
    account_value = (
        signal_data
        .select("date", "asset", "prev_close", "limit_qty")
        .unique(subset=["date", "asset"])
        .group_by("date")
        .agg(
            (pl.col("prev_close") * pl.col("limit_qty")).sum().alias("account_value")
        )
    )
    daily_asset_num = (
        signal_data
        .group_by("date")
        .agg(asset_num=pl.col("asset").n_unique())
    )

    # todo: summary - 交易胜率/日胜率/平均单笔/盈亏比/股票覆盖度/信号量/换手/ 预测胜率 / 交易胜率
    day_pnl_df = (
        order_uc
        .group_by("date")
        .agg(pl.col("pnl").sum())
        .join(account_value, on="date", how="left")
        .select("date", "pnl", 'account_value')
        .sort("date")
    )
    # 交易胜率
    win_ratio = order_uc.select((pl.col("pnl") > 0).mean())
    # 日胜率
    day_win_ratio = day_pnl_df.select((pl.col("pnl") > 0).mean())
    # 平均单笔
    bp = day_pnl_df.select((pl.col("pnl").sum() / pl.col("account_value").sum() * 1e4))
    # 盈亏比
    win_loss_ratio = (
            order_uc
            .filter(pl.col("pnl") > 0)
            .select("pnl")
            .sum() /
            order_uc
            .filter(pl.col("pnl") < 0)
            .select(pl.col("pnl").abs())
            .sum()
    )
    summary_df = pl.concat([win_ratio,
                            day_win_ratio,
                            bp,
                            win_loss_ratio],
                           how="vertical")
    summary_df = summary_df.to_pandas()
    summary_df.index = pd.Index(["Win Ratio/Trade", "Win Ratio/Day", "BP", "Profit/Loss"], name="metric")
    # summary_df = dc.to_pl(summary_df.reset_index()).select("metric", *[pl.col(period).round(2) for period in periods_cols])

    # print(summary_df)
    tb = finplot.table(summary_df.round(2).reset_index())
    if show_notebook:
        tb.show()
    # 股票覆盖度
    coverage = (
        order_uc
        .group_by("date")
        .agg(pl.col("asset").n_unique().alias("signal_asset_num"))
        .join(daily_asset_num, on="date", how="right", )
        .select("date", coverage=(pl.col("signal_asset_num") / pl.col("asset_num")).fill_null(0.0))
        .sort("date")
    )

    # 换手
    turnover = (
        order_uc
        .group_by("date")
        .agg(open_amt=(pl.col("order_qty") * pl.col("price")).sum())
        .join(account_value, on="date", how="right")
        .select("date", turnover=pl.col("open_amt") / pl.col("account_value"))
        .fill_null(0.0)
        .sort("date")
    )
    fig_coverage = finplot.bar_plot(coverage.to_pandas().set_index("date", drop=True),
                                    title=f"coverage (median: {(coverage["coverage"].median() * 100):.2f}%)")
    if show_notebook:
        fig_coverage.show()
    fig_turnover = finplot.bar_plot(turnover.to_pandas().set_index("date", drop=True),
                                    title=f"turnover (median: {(turnover["turnover"].median() * 100):.2f}%)")
    if show_notebook:
        fig_turnover.show()
    figs = [tb, fig_coverage, fig_turnover]
    nv_df = (
        day_pnl_df
        .with_columns(pl.col("pnl") / pl.col("account_value"))
        .fill_null(0.0)
        .with_columns((pl.col("pnl") + 1).cum_prod())
    )
    nv_pd = nv_df.to_pandas().set_index("date", drop=True)
    fig, metric = finplot.nv_plot(nv_pd["pnl"])

    summary = dc.to_pl(summary_df.reset_index()).select("metric", pl.col("pnl").round(2))
    summary_dict = summary.to_dicts()
    for item in summary_dict:
        metric[item["metric"]] = item["pnl"]
    if show_notebook:
        fig.show()
    figs.append(fig)

    return {
        "info": order_uc,
        "summary": summary,
        "pnl": day_pnl_df,
        "nv": nv_df,
        "figs": figs,
    }, metric
